import sys

from Organisms.GA.Fitness_Operators.Fitness_Operator import Fitness_Operator, dot
from Organisms.GA.Fitness_Operators.SCM_and_Energy_Fitness_Operator_Options import SCM_options

from Organisms.GA.Fitness_Operators.Energetic_Fitness_Contribution import get_lowest_and_highest_energies_from_collections
from Organisms.GA.Fitness_Operators.CNA_Fitness_Contribution       import get_lowest_and_highest_similarities_from_collections

from Organisms.GA.Fitness_Operators.Energetic_Fitness_Contribution import get_energetic_fitness_contribution, get_lowest_and_highest_energies_from_collections
from Organisms.GA.Fitness_Operators.CNA_Fitness_Contribution       import get_lowest_and_highest_similarities_from_collections

from Organisms.GA.Fitness_Operators.CNA_Fitness_Contribution import get_CNA_most_similar_average

def get_average(cluster_SCM_simiarities):
	average = float(sum(cluster_SCM_simiarities)) / float(len(cluster_SCM_simiarities))
	return average

class SCM_and_Energy_Fitness_Operator(Fitness_Operator):
	"""
	This class controls the how the fitness values for the Structural Comparison Method (SRM)-based predation operator are obtained.
	
	:param fitness_information: This is all the information that is needed about the fitness class
	:type  fitness_information: dict.
	:param predation_operator: This is the predation operator that this fitness class will take information from if needed to obtain a fitness.
	:type  predation_operator: Organisms.GA.Predation_Operator
	:param population: The population to assign fitnesses to
	:type  population: Organisms.GA.Population
	:param generations: The number of generations that will be performed in the genetic algorithm.
	:type  generations: int
	:param no_of_cpus: the number of cpus to use when multiprocessing
	:type  no_of_cpus: int
	:param print_details: Print the details of the energy fitness operator. True if yes, False if no
	:type  print_details: bool

	"""
	def __init__(self,fitness_information, predation_operator, population, generations, no_of_cpus, print_details): # Check this out.
		super().__init__(fitness_information, population, print_details)
		SCM_options(self, fitness_information, predation_operator, no_of_cpus)

	def print_initial_message(self):
		"""
		At the start of the GA run, this message will be shown so that the user knows what their energy and fSRM (CNA)-based itness contributions are at the start of the GA.
		"""
		print('****************************************************')
		print('The Fitness Contribution from the Energy and CNA components.')
		print('Energetic Contribution to Fitness: '+str(self.energy_contribution))
		print('CNA Contribution to Fitness: '+str(self.CNA_contribution))
		print('****************************************************')
		
	####################################################################################
	################## Method for assign_initial_population_fitnesses ##################
	####################################################################################

	def assign_initial_population_fitnesses(self): 
		"""
		This method is designed to assign fitness values to the clusters of the population only at the start of the GA, when the population has been initialised

		"""
		self.cna_database.add(self.population,initialise=True)
		self.__assign_population_fitnesses(0)

	####################################################################################
	################## Method for assign_resumed_population_fitnesses ##################
	####################################################################################

	def assign_resumed_population_fitnesses(self,resume_from_generation): 
		"""
		This method is designed to assign fitness values to the clusters of the population only at the beginning of a resumed GA. 

		:param resume_from_generation: The current generation to resume from.
		:type  resume_from_generation: int
		
		"""
		self.cna_database.add(self.population,initialise=True)
		self.__assign_population_fitnesses(resume_from_generation)

	############################################################################################################
	################## Method for assign_all_fitnesses_before_assess_against_predation_operator ##################
	############################################################################################################

	def assign_all_fitnesses_before_assess_against_predation_operator(self,all_offspring_pools,current_generation_no):
		"""
		This method is to be used in the GA program. 
		This will assign all the fitnesses of all clusters in the current generation (population and offspring) before the offspring are assessed to understand if they violate the predation operator (i.e. an offspring is Class 1 similar to a cluster in the population, or another offspring).

		See the description given for the "assign_all_fitnesses" def on what the crux of this method is.

		If you write your own diversiy and fitness classes, you do not need to implement this method in your fitness class.

		:param all_offspring_pools: list of all the offspring pools of offspring to assign fitness values to.
		:type  all_offspring_pools: list of Organisms.GA.Offspring_Pool
		:param current_generation_no: The current generation
		:type  current_generation_no: int
		"""
		if not isinstance(all_offspring_pools,list):
			all_offspring_pools = [all_offspring_pools]
		#for offspring_pool in all_offspring_pools:
		#	self.cna_database.add(offspring_pool,initialise=False)
		return self.__assign_all_fitnesses(all_offspring_pools,current_generation_no)

	###########################################################################################################
	################# Method for assign_all_fitnesses_after_assess_against_predation_operator #################
	###########################################################################################################

	def assign_all_fitnesses_after_assess_against_predation_operator(self,all_offspring_pools,current_generation_no,offspring_to_remove):
		"""
		This method is to be used in the GA program. 
		This will assign all the fitnesses of all clusters in the current generation (population and offspring) after the offspring are assessed to understand if they violate the predation operator (i.e. an offspring is Class 1 similar to a cluster in the population, or another offspring).

		See the description given for the "assign_all_fitnesses" def on what the crux of this method is.

		If you write your own diversiy and fitness classes, you do not need to implement this method in your fitness class.

		:param all_offspring_pools: All of the offspring_pools
		:type  all_offspring_pools: list of Organisms.GA.Offspring_Pool
		:param current_generation_no: The current generation
		:type  current_generation_no: int
		:param offspring_to_remove: a list of all the names of clusters in the offspring that will be removed. Currently not needed, but kept in case. 
		:type  offspring_to_remove: list of int

		"""
		# If self.use_same_CNA_database == False, then the first thing that needs to be done is 
		# to update self.cna_database to remove clusters that would have been removed during 
		# self.predation_operator.remove_offspring_and_replace_with_population_that_violate_predation_operator
		# when the generations are being performed in the GA_Program script.
		#if self.use_same_CNA_database == False:
		#	offspring_to_remove_names = [name for name,index in offspring_to_remove]
		#	self.cna_database.remove(offspring_to_remove_names)
		return self.__assign_all_fitnesses(all_offspring_pools,current_generation_no)

	#############################################################################################
	################## Method for assign_all_fitnesses_after_natural_selection ##################
	#############################################################################################

	def assign_all_fitnesses_after_natural_selection(self,current_generation_no):
		"""
		This method is to be used in the GA program. 
		This will assign all the fitnesses to all the clusters in the population before the offspring are assessed to understand if they violate the predation operator (i.e. an offspring is Class 1 similar to a cluster in the population, or another offspring).

		See the description given for the "assign_population_fitnesses" def on what the crux of this method is.

		If you write your own diversiy and fitness classes, you do not need to implement this method in your fitness class.

		:param current_generation_no: The current generation
		:type  current_generation_no: int

		"""
		# Remove all clusters that were not included in the population due to the natural selection process.
		'''
		clusters_to_remove_from_cna_database = []
		clusters_in_population = self.population.get_cluster_names()
		for cluster_name in self.cna_database.keys():
			if not cluster_name in clusters_in_population:
				clusters_to_remove_from_cna_database.append(cluster_name)
		'''
		#self.cna_database.remove(clusters_to_remove_from_cna_database)
		# Just doing a check to make sure that the number of entries in the self.cna_database is being reduced after every generation, preventing memory leaking
		#self.cna_database.check_database(self.population)
		# Assign fitnesses
		cluster_SCM_simiarities = self.__assign_population_fitnesses(current_generation_no)
		self.__update_weights(cluster_SCM_simiarities)

	def __update_weights(self,cluster_SCM_simiarities):
		"""
		Updates the energy and SCM coefficients if the dynamic version of the structure + energy fitness operator is used.

		:param cluster_SCM_simiarities: The similarities of clusters of the population as obtained by the SCM. These are the sigma 1/2 similarities. 
		:type  cluster_SCM_simiarities: list of floats
		"""
		if self.dynamic_mode:
			# Obtain the population 'similarity value' from all the similarity values from all clusters in the population. 
			population_fitness = self.get_population_fitness(cluster_SCM_simiarities,self.population_fitness_function)
			# The population 'fitness value' is used to update the coefficients (weights) for the energy and SCM coefficients.
			self.SCM_fitness_contribution = self.convert_population_fitness_to_SCM_fitness_contribution(population_fitness)
			# obtain the new coefficients for energy, and place them into self.fitness_weights. The format of self.fitness_weights must be [energy_fitness_coefficient, SCM_fitness_coefficient]
			self.energy_fitness_contribution = 1.0 - self.SCM_fitness_contribution
			self.fitness_weights = []
			if not self.energy_fitness_contribution == 0.0:
				self.fitness_weights.append(self.energy_fitness_contribution)
			if not self.SCM_fitness_contribution == 0.0:
				self.fitness_weights.append(self.SCM_fitness_contribution)

	def get_population_fitness(self,cluster_SCM_simiarities,population_fitness_function):
		"""
		Obtain the 'fitness' of the population. This population fitness is based on the similarity values from all clusters in the population. 
		This python definition will take the similarity values from all clusters in the population and get a 'similarity value' of the population. 
		This definition will then take this population similarity value and convert it into a population fitness value. 
		The population fitness value must be a read value between (and including) 0.0 and 1.0.

		:param cluster_SCM_simiarities: The similarities of clusters of the population as obtained by the SCM. These are the sigma 1/2 similarities. 
		:type  cluster_SCM_simiarities: list of floats
		:param population_fitness_function: The python def that turns the population similarity value into a population fitness value.
		:type  population_fitness_function: def

		"""
		# THis is just an example which geoff is having a mess around with
		# from https://pdf.sciencedirectassets.com/280203/1-s2.0-S1877050918X00076/1-s2.0-S1877050918306100/main.pdf?X-Amz-Security-Token=IQoJb3JpZ2luX2VjEPn%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCXVzLWVhc3QtMSJGMEQCIGj3e8QchiulQhGRvnjboT7FpZqiI27yQRKXZruql7etAiA8cKouDjyTz0Z4dE%2BuPF09Jh7j3TyRtcEI3pjGm%2BEvayq0AwhyEAMaDDA1OTAwMzU0Njg2NSIMKfbjirK0jaEfegOBKpEDFTS0BzXwH84wlrINvGrWjj%2B7iae1aUvWuZHY0HK0DBqyNUC8Ws9mX3zMw9pD96tiIH4ER921DG5eCY8i%2Byv1aqEiO5ELgt4Ny0%2B9nLJXIvINi3ajOvXqhkygomRmQPwRH9fHm9tH6x5c2RG0JQerIMG0yalE1%2BlYO%2FwdizaeCUruvug32x0n%2B2pfqc32zjIComIh%2B0WhomLSgijxAQhy4Q7n6flosjb9i5k6TaZQ5Z6jHReNsNZBXL06SiA5wGtGirgW%2F5dGoRIceP6paAP6cAi03eoBzpFefp556hcAQO8VAlFVybITXJr72i%2BqAQJhQnZGKu4uu1dz%2FRiCRq63qWhJlhLrWgHBgYJglwK0Km7s2zB58be7FWARekdjlXWucPug4den3hp%2FLgh4fB%2Bn%2FmJUWNl2z2WUWh%2FgBhrbw%2FZ7Wmv9N9SCXW2LXAjuCdaD6FMFQuM96FpxhmamMFggzd41H3jYINrJ%2F4D5lyI6WC9lMiVnG%2B6BC1ywdc6tlD%2BjSWLaB0p8%2F%2FHUZqKKTZd%2FRFMw%2F4Pe%2FQU67AE4oTrypN69ZmxuFP%2FDJ5QjDIL4ZdNGuFxoBP46SoDn44Z%2Fsp%2Bofib8SFZCkIo3Peb0rRKE01atFcddH9j3TN8Ft8muloLoPlV4RJu14VPZtwpTt%2BzRth5I1LjtJSw2FVVeEuAb%2BAO3YmlMUc7bRzLDe3dLghSIDLXArz3VhQrAB0vMmWQUbmSoF34B89AjNcv3mYh53jJAzYpx7n8d%2B2mt2xl%2FH1J4DHvmoyn6syHA4MdTiCugNCnLBmGqN1bVn9S6bCe349dftzdRwCoBjDanZ2l0CMhMiK5ZV1bni0vK1WCPR9FQpQRxyPDo3A%3D%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20201120T094441Z&X-Amz-SignedHeaders=host&X-Amz-Expires=300&X-Amz-Credential=ASIAQ3PHCVTYRCMUJEVX%2F20201120%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Signature=b6d1076269a5e78baa96eb7767b830747ebd951f54c72663b69a906be8d1100a&hash=4deb11b77677cd260b84379bebb2c3fcb2bcab1036d6a3c1d95df37b83c2d45b&host=68042c943591013ac2b2430a89b270f6af2c76d8dfd086a07176afe7c76c2c61&pii=S1877050918306100&tid=spdf-fdbaed26-450e-41d4-9f78-2f601327c053&sid=98c7a6d02333b5488f4aa07-ffcec9794cf8gxrqa&type=client
		average_similarity_value = get_average(cluster_SCM_simiarities)
		min_similarity_value = min(cluster_SCM_simiarities)
		max_similarity_value = max(cluster_SCM_simiarities)
		diff_similarity_value = max_similarity_value - min_similarity_value
		k_value = 1.0
		population_similarity_value = (1.0 - ((average_similarity_value - min_similarity_value)/diff_similarity_value)) ** k_value
		#population_fitness_value = population_fitness_function(population_similarity_value)
		population_fitness_value = population_similarity_value
		return population_fitness_value

	def convert_population_fitness_to_SCM_fitness_contribution(population_fitness):
		'''
		This method will take the population fitness and use this to determine what the new coefficients (weights) for the energy and SCM coefficients are.

		:param population_fitness: The population 'fitness value'. This value is some value between 0.0 and 1.0.
		:type  population_fitness: float

		'''
		return population_fitness

	#####################################################################################
	################## Subsidiary Methods for the Energy Fitness Class ##################
	#####################################################################################

	def __assign_population_fitnesses(self,current_generation): 
		"""
		This method is designed to assign fitness values to the clusters of the population only. 

		This methos also returns the cluster_SCM_simiarities which the algorithm then uses to update the c_e and c_SCM values. 

		:param current_generation_no: The current generation.
		:type  current_generation_no: int

		:returns: The similarities of the clusters. These are he sigma 1/2 similarities. 
		:rtype:   list of floats
		
		"""
		# Update the maximum and minimum energy of the population.
		lowest_energy, highest_energy = get_lowest_and_highest_energies_from_collections(self.population, None)
		# Update the maximum and minimum similarities of the population.
		if self.normalise_similarities:
			lowest_similarity, highest_similarity = get_lowest_and_highest_similarities_from_collections(self.population, None, self.cna_database, self.collection_function)
		else:
			lowest_similarity, highest_similarity = None, None
		# Update the fitness of all clusters in the population
		cluster_SCM_simiarities = self.__update_collection_fitnesses(self.population, highest_energy, lowest_energy, highest_similarity, lowest_similarity)
		return cluster_SCM_simiarities

	def __assign_all_fitnesses(self,all_offspring_pools,current_generation): 
		"""
		Update the fitnesses of all clusters in population and all offsprings

		:param all_offspring_pools: All of the offspring_pools
		:type  all_offspring_pools: list of Organisms.GA.Offspring_Pool
		:param current_generation_no: The current generation
		:type  current_generation_no: int

		"""
		if not isinstance(all_offspring_pools,list):
			all_offspring_pools = [all_offspring_pools]
		# Update the maximum and minimum energy of the population and offsprings as a whole.
		lowest_energy, highest_energy = get_lowest_and_highest_energies_from_collections(self.population, all_offspring_pools)
		# Update the maximum and minimum similarities of the population and offsprings as a whole.
		if self.normalise_similarities:
			lowest_similarity, highest_similarity = get_lowest_and_highest_similarities_from_collections(self.population, all_offspring_pools, self.cna_database, self.collection_function)
		else:
			lowest_similarity, highest_similarity = None, None	
		# Update the fitness of all clusters in the population
		cluster_SCM_simiarities = self.__update_collection_fitnesses(self.population, highest_energy, lowest_energy, highest_similarity, lowest_similarity)
		# Update the fitness of all clusters in all the offsprings
		for offspring_pool in all_offspring_pools:
			self.__update_collection_fitnesses(offspring_pool, highest_energy, lowest_energy, highest_similarity, lowest_similarity)

	def __update_collection_fitnesses(self,collection,highest_energy,lowest_energy,highest_similarity,lowest_similarity):
		"""
		The fitness of all the clusters in the collection are updated. The collection could be the population or the offspring_pool

		:param collection: update the fitnesses of clusters in the collection.
		:type  collection: Organisms.GA.Collection.Collection
		:param highest_energy: The highest energy of the clusters in the collection. 
		:type  highest_energy: float
		:param lowest_energy: The lowest energy of the clusters in the collection. 
		:type  lowest_energy: float

		:returns: The similarities of the clusters. These are he sigma 1/2 similarities. 
		:rtype:   list of floats

		"""
		cluster_SCM_simiarities = []
		for cluster in collection:
			fitness, cluster_SCM_simiarity = self.__get_fitness(cluster, highest_energy, lowest_energy, highest_similarity, lowest_similarity)
			cluster.fitness = fitness
			cluster_SCM_simiarities.append(cluster_SCM_simiarity)
		#import pdb; pdb.set_trace()
		return cluster_SCM_simiarities

	def __get_fitness(self,cluster,highest_energy,lowest_energy,highest_similarity,lowest_similarity):
		"""
		Get the fitness of a cluster.

		:param cluster: The cluster to get the fitness of
		:type  cluster: Organisms.GA.Cluster.Cluster
		:param highest_energy: The highest energy of the clusters in the collection. 
		:type  highest_energy: float
		:param lowest_energy: The lowest energy of the clusters in the collection. 
		:type  lowest_energy: float

		:returns: The similarities of the clusters. These are he sigma 1/2 similarities. 
		:rtype:   list of floats

		"""
		fitness_contributions = []
		# get fitness in terms of energy
		if (not self.energy_fitness_contribution == 0.0):
			energy_fitness_contribution = get_energetic_fitness_contribution(cluster,highest_energy,lowest_energy,self.energy_fitness_function)
			fitness_contributions.append(energy_fitness_contribution)
		# get fitness in terms of CNA predation
		if (not self.SCM_fitness_contribution == 0.0) or self.dynamic_mode:
			CNA_fitness_contribution, cluster_SCM_simiarity = self.get_CNA_fitness_contribution(cluster,highest_similarity,lowest_similarity,self.cna_database,self.collection_function,self.cna_fitness_function)
			fitness_contributions.append(CNA_fitness_contribution)
		else:
			cluster_SCM_simiarity = None # Dont need this if self.SCM_fitness_contribution == 0.0 or self.dynamic_mode == False
		# Get the overall fitness
		overall_fitness = dot(self.fitness_weights,fitness_contributions)
		return overall_fitness, cluster_SCM_simiarity

	def add_to_database(self, collection):
		'''
		Add clusters similarities to the CNA database to be stored for future generations. 

		:param collection: update the fitnesses of clusters in the collection.
		:type  collection: Organisms.GA.Collection.Collection
		'''
		self.cna_database.add(collection,initialise=False)

	def remove_from_database(self, cluster_names_to_remove):
		'''
		Clusters to remove from the CNA database

		:param cluster_names_to_remove: A list of the names of all the clusters to remove from the CNA database
		:type  cluster_names_to_remove: list of ints
		'''
		self.cna_database.remove(cluster_names_to_remove) # delete the cluster entrance in the self.cna_database

	def reset(self):
		'''
		Reset the CNA database with no inputs
		'''
		self.cna_database.reset()

	def is_there_an_similarity_range(self,similarity_rounding):
		"""
		Determines if there is a range of similarities in the collection

		:param rounding: The rounding of the similarity of the cluster
		:type  rounding: float

		returns Is there a range of similarities in the collection
		rtype   bool 
		"""
		cluster_similarities = [get_CNA_most_similar_average(cluster,self.cna_database,self.collection_function) for cluster in self.population]
		if round(max(cluster_similarities),similarity_rounding) == round(min(cluster_similarities),similarity_rounding):
			return False
		else:
			return True

	# ------------------------------------------------------------------------------------------------------------------------------------------------------




'''
Recording_Clusters.py, 02/10/2018, Geoffrey R Weal

'''
import sys
from math import pi
import numpy as np

from Organisms.GA.Collection import Collection
from Organisms.GA.GA_Program_external_methods import update_progress

class GA_Recording_Database(Collection):
	"""
	This is a Collection that has been designed to record clusters that have been created during the genetic algorithm.

	This has been designed to give the user many ways to limit the number of clusters that are recorded. This is to prevent the size of the database from getting to big in disk size.
	
	:param path_of_database: The path to this database
	:type  path_of_database: str.
	:param max_no_of_recorded_structures: This is the maximum number of clusters that will be recorded. If this limit is reached, the higher energy clusters will be replaced in new, lower energy clusters.
	:type  max_no_of_recorded_structures: int
	:param limit_datasize_of_database: This is the maximum size the database can get. Give as a string with size + memory type (e.g. 150MB, 2.0GB)
	:type  limit_datasize_of_database: str.
	:param ga_recording_scheme: The user can indicate a specific type of recording scheme to use to limit the clusters that are recorded. See manual for more details. Default: 'None'
	:type  ga_recording_scheme: str.
	:param limit_energy_height_of_clusters_recorded: If ga_recording_scheme == 'Limit_energy_height': is selected, this is the maximum energy above the LES energy. Any clusters that are lower in energy than Energy(LES) + limit_energy_height_of_clusters_recorded will be recorded. Any that have an energy higher than this will not be recorded. Default: float('inf')
	:type  limit_energy_height_of_clusters_recorded: float
	:param lower_energy_limit: If ga_recording_scheme == Set_energy_limits is selected, this is the low energy limit. Any cluster with an energy lower than lower_energy_limit will not be recorded. Default: -float('inf')
	:type  lower_energy_limit: float
	:param upper_energy_limit: If ga_recording_scheme == Set_energy_limits or ga_recording_scheme == 'Set_higher_limit', this is the upper energy limit. Any clusters with an energy higher than this energy limit will not be recorded. Default: float('inf')
	:type  upper_energy_limit: float

	"""
	def __init__(self,path_of_database,max_no_of_recorded_structures=None,limit_datasize_of_database=None,ga_recording_scheme='None',limit_energy_height_of_clusters_recorded=float('inf'),lower_energy_limit=-float('inf'),upper_energy_limit=float('inf'),show_GA_Recording_Database_check_percentage=False): 
		self.limit_datasize_of_database = limit_datasize_of_database
		self.ga_recording_scheme = ga_recording_scheme
		# Get the setting for any of the ga_recording_scheme
		if self.ga_recording_scheme == 'All':
			self.write_initial_cluster_in_RAM = False
			self.write_cluster_in_RAM = False
		elif self.ga_recording_scheme == 'Limit_energy_height':
			self.limit_energy_height_of_clusters_recorded = limit_energy_height_of_clusters_recorded
			self.write_initial_cluster_in_RAM = True
			self.write_cluster_in_RAM = True
		elif self.ga_recording_scheme == 'Set_higher_limit':
			self.upper_energy_limit = upper_energy_limit
			self.write_initial_cluster_in_RAM = False
			self.write_cluster_in_RAM = True
		elif self.ga_recording_scheme == 'Set_energy_limits':
			self.lower_energy_limit = lower_energy_limit
			self.upper_energy_limit = upper_energy_limit
			self.write_initial_cluster_in_RAM = False
			self.write_cluster_in_RAM = True
		elif self.ga_recording_scheme == 'None':
			exit('Error, should not create a GA_Recording_Database if self.ga_recording_scheme == None')
		# Get the settings for if self.size or self.limit_datasize_of_database is not equal to none
		if (max_no_of_recorded_structures == None) or (self.limit_datasize_of_database == None):
			self.write_initial_cluster_in_RAM = True
			self.write_cluster_in_RAM = True
		# Show the GA recording database check percentage
		self.show_GA_Recording_Database_check_percentage = show_GA_Recording_Database_check_percentage
		# Write the rest of the collection
		name = 'GA_Recording_Database'
		Collection.__init__(self,name,max_no_of_recorded_structures,have_database=True,path=path_of_database,write_collection_history=False,write_cluster_in_RAM=self.write_cluster_in_RAM)

	# ----------------------------------------------------------------------------------------------------------------------

	def get_cluster_names(self,order=False):
		"""
		Will provide a list of all the names of all the clusters in the Collection

		Inputs:
			order (bool.): This tag will tell this method whether the user would like the list of names given in order. 

		Returns:
			List of the names of all the clusters in the Population
		"""
		if self.write_cluster_in_RAM:
			cluster_names = [cluster[0] for cluster in self.clusters]
		else:
			import pdb; pdb.set_trace()
			exit('Cant do this')
		if order == True:
			cluster_names.sort()
		return cluster_names

	def sort_by_name(self):
		"""
		This method will sort the clusters in the list by their name.
		"""
		if self.write_cluster_in_RAM:
			self.clusters.sort(key=lambda cluster: cluster[0])
		else:
			import pdb; pdb.set_trace()
			exit('Cant do this')

	def sort_by_energy(self):
		"""
		This method will sort the clusters in the list by their energy (from lowest energy to highest energy).  
		"""
		if self.write_cluster_in_RAM:
			self.clusters.sort(key=lambda cluster: cluster[1])
		else:
			import pdb; pdb.set_trace()
			exit('Cant do this')

	def get_index(self,name_to_find):
		"""
		This method will provide the index of the cluster that has the name "name_to_find" in the Collection

		Inputs:
			name_to_find (int): the name of the cluster in the Collection to obtain the index for

		Returns:
			the index of the cluster in the Collection with the name "name_to_find"

		Exceptions:
			Will break if the cluster with the name "name_to_find" can not be found in this method.
		"""
		if self.write_cluster_in_RAM:
			for index in range(len(self.clusters)):
				if self.clusters[index][0] == name_to_find:
					return index
		else:
			print('Error in def get_index in class GA_Recording_System, in GA_Recording_System.py')
			print('You can use this method as the collection is not recording clusters in the RAM')
			print('Check your setting for write_cluster_in_RAM for '+str(self))
			exit('This program will exit without completing')
		print('Error in def get_index, in Class GA_Recording_Database, in GA_Recording_System.py')
		print('Error in def get_index in GA_Recording_System.py: Could not find the cluster that has the name = '+str(name_to_find))
		print('Check this.')
		exit('This program will exit without completing.')

	def add(self, index, cluster):
		"""
		Adds a cluster to the Collection.

		Index:
			index (int/str.): the index of the ith cluster in the Collection. If "End" is inputed, the cluster will be append to the end of the Collection list.
			cluster (Organisms.GA.Cluster): The cluster to add at the ith position in the Collection. 
		"""
		# Add the cluster to the Collection
		if self.write_cluster_in_RAM:	
			if index == 'End':
				self.clusters.append((cluster.name,cluster.energy))
			else:
				self.clusters.insert(index,(cluster.name,cluster.energy))
		else:
			self.no_of_clusters_counter += 1
		# adds the cluster in the collection to disk
		if self.have_database:
			self.add_to_database(cluster,center=True)

	def remove_to_database(self,cluster):
		"""
		Allows the user to remove a cluster in the collection from the  ASE database

		Inputs:
			cluster (Organisms.GA.Cluster): The cluster to remove from the database.
		"""
		with connect(self.database_path) as database:
			cluster_id_to_remove = database.get(name=cluster[0]).id
			del database[cluster_id_to_remove]

	def add_clusters_into_RAM(self, cluster_dict, cluster_names):
		"""
		This method adds clusters into the RAM

		:param cluster_dict: This is a dicionary of all the clusters from the database, given as {cluster_name: Cluster}
		:type  cluster_dict: {int: ASE.Cluster}
		:param cluster_names: list of the names of the clusters that are needed for the collection
		:type  cluster_names: list of int
		"""
		if self.write_cluster_in_RAM:	
			for cluster_name in cluster_names:
				cluster = cluster_dict[cluster_name]
				self.clusters.append((cluster.name,cluster.energy))

	# ----------------------------------------------------------------------------------------------------------------------
		
	def add_collection_to_database(self, collection, clusters_not_to_include):
		"""
		Will record the clusters in the collection to the GA_Recording_Database. 

		:param collection: The collection to be recorded
		:type  collection: Organisms.GA.Collection
		:param clusters_not_to_include: A list of names of clusters in the collection not to record in GA_Recording_Database.
		:type  clusters_not_to_include: list of int
		"""
		if self.ga_recording_scheme == 'None':
			return 
		# Add all clusters that you have in the collection to the GA recording database
		elif self.ga_recording_scheme == 'All':
			for cluster in collection:
				if not cluster.name in clusters_not_to_include:
					self.add(-1,cluster)
			#self.sort_by_energy()
		# Add all clusters that you have in the collection to the GA recording database
		elif self.ga_recording_scheme == 'Limit_energy_height': 
			for cluster in collection:
				if not cluster.name in clusters_not_to_include:
					self.add(-1,cluster)
			self.sort_by_energy()
			for index in range(len(self)-1,-1,-1): 
				if self[index][1] > (self[0][1] + self.limit_energy_height_of_clusters_recorded):
					del self[index]
				else:
					break
		# Add all clusters that you have in the collection to the GA recording database
		elif self.ga_recording_scheme == 'Set_higher_limit': 
			for cluster in collection:
				if (cluster.energy < self.upper_energy_limit) and (not cluster.name in clusters_not_to_include):
					self.add(-1,cluster)
		# Add all clusters that you have in the collection to the GA recording database
		elif self.ga_recording_scheme == 'Set_energy_limits': 
			for cluster in collection:
				if (self.lower_energy_limit < cluster.energy < self.upper_energy_limit) and (not cluster.name in clusters_not_to_include):
					self.add(-1,cluster)
		# get rid of all the highest energy clusters that can not fit into the database
		if (not self.size == None) and (len(self) > self.size):
			self.sort_by_energy()
			while len(self) > self.size: # of if size is greater than current size 
				del self[-1] 
		if (not self.limit_datasize_of_database == None) and (os.stat(self.database_path).st_size > self.limit_datasize_of_database):
			self.sort_by_energy()
			while os.stat(self.database_path).st_size > self.limit_datasize_of_database: # of if size is greater than current size
				del self[-1]
		#database.metadata = {'title': str(name_of_database), 'key_descriptions': {'systems': ('Cluster', 'The cluster', ''), 'gen_made': ('Generation Created', 'The generation tthe cluster was created.', ''), 'cluster_energy': ('Cluster Energy', 'Potential energy of the cluster.', 'eV'), 'name': ('Name', 'Name of the cluster.', '')}, 'default_columns': ['name', 'cluster_energy', 'id']}

	def update_cluster_in_database_for_if_in_population(self, clusters_in_the_population, energies_of_clusters_removed_from_the_population):
		"""
		This method will update clusters in the database if that cluster was ever accepted into the population. 

		:param clusters_in_the_population: This is a list of the names of clusters that are in the population.
		:type  clusters_in_the_population: list of int
		"""
		def names_of_clusters_in_database(database):
			for row in database.select():
				yield row.name

		if self.ga_recording_scheme == 'All':
			with connect(self.database_path) as database:
				for name in clusters_in_the_population:
					row = database.update(name,ever_in_population=True)
		elif self.ga_recording_scheme == 'Limit_energy_height': 
			with connect(self.database_path) as database:
				for name in clusters_in_the_population:
					if name in names_of_clusters_in_database(database):
						row = database.update(name,ever_in_population=True)
		elif self.ga_recording_scheme == 'Set_higher_limit': 
			with connect(self.database_path) as database:
				for name, energy in zip(clusters_in_the_population,energies_of_clusters_removed_from_the_population):
					if energy <= self.upper_energy_limit:
						row = database.update(name,ever_in_population=True)
		elif self.ga_recording_scheme == 'Set_energy_limits': 
			with connect(self.database_path) as database:
				for name, energy in zip(clusters_in_the_population,energies_of_clusters_removed_from_the_population):
					if self.lower_energy_limit <= energy <= self.upper_energy_limit:
						row = database.update(name,ever_in_population=True)
						
	# -----------------------------------------------------------------------------------------------------
	# Methods for restoring the ga_recording_database from a particular generation. 

	def check_clusters_in_database(self,generation):
		"""
		Will check the clusters in the database and remove any cluster that was created during a most recent unsuccessful generation.

		:param generation: The current generation
		:type  generation: int

		"""
		# remove any files that will prevent it from being read.
		if os.path.exists(self.database_path+'.lock'):
			os.remove(self.database_path+'.lock')
		if os.path.exists(self.database_path+'-journal'):
			os.remove(self.database_path+'-journal')
		# print information about the progress bar loading
		if self.show_GA_Recording_Database_check_percentage:
			sys.stderr.write('The progress bar below shows the progress of reading and checking the GA Recording System database.\n')
		ids_to_remove = []
		# check clusters from the database
		with connect(self.database_path) as database:
			total_size_of_database = float(database.count())
			if self.show_GA_Recording_Database_check_percentage:
				counter = 0.0;
			# Get the last id from the database. The algorithm will cycle through from id=1 to id=last_id
			with database.managed_connection() as con: 
				cur = con.cursor()
				last_id = database.get_last_id(cur)
			# Check every cluster in the database
			for an_id in range(last_id):
				try:
					row = database.get(selection=an_id,columns=['id','key_value_pairs'],include_data=False,verbosity=0)
				except KeyError:
					continue
				if row.gen_made > generation:
					ids_to_remove.append(an_id)
				if self.show_GA_Recording_Database_check_percentage:
					counter += 1.0
					current_percentage = counter/total_size_of_database
					update_progress(current_percentage)
			# remove clusters that were created in the lastest non successful generation.
			database.delete(ids_to_remove)
			if self.show_GA_Recording_Database_check_percentage:
				update_progress(1.0)
		'''
		offset = 0; offset_max = 100; counter = 0.0;
		got_to_end_of_database = False
		with connect(self.database_path) as database:
			total_size_of_database = float(database.count())
		while not got_to_end_of_database:
			counter = 0
			with connect(self.database_path) as database:
				for row in database.select(columns=['id','key_value_pairs'],include_data=False,offset=offset,verbosity=0):
					if row.gen_made > generation:
						ids_to_remove.append(row.id)
					counter += 1.0
					current_percentage = counter/total_size_of_database
					update_progress(current_percentage)
				if counter < offset_max:
					offset += offset_max
				else:
					database.delete(ids_to_remove)
					got_to_end_of_database = True
					update_progress(1.0)
		'''

	def import_information_from_database(self,current_generation):
		"""
		Import any data from the database if it is needed. This should only be needed for the 'Limit_energy_height' scheme.

		:param current_generation: The current generation that your genetic algorithm trial is being resumed from
		:type  current_generation: int

		"""
		if self.write_initial_cluster_in_RAM:
			return 
		with connect(self.database_path) as database:
			for row in database.select():
				self.clusters.append((row.name,row.energy))
		self.sort_by_energy()

# ----------------------------------------------------------------------------------------------------------------------------

import os, sys
from shutil import rmtree, copytree, copyfile
from ase.db import connect

def get_size(size):
    """
	will convert the most human friendly version of the disk space of the database. 

	Input:
		size (float): the disk space of the database in bytes
    """
    power = 2**10
    n = 0
    Dic_powerN = {0 : '', 1: 'kilo', 2: 'mega', 3: 'giga', 4: 'tera'}
    while size > power:
        size /=  power
        n += 1
    return size, Dic_powerN[n]+'bytes'

def make_folder(path_to_folder):
	"""
	Will remake a folder, even if it already exists.

	Input:
		path_to_folder (str.): the path to the folder to remake.
	"""
	if not os.path.exists(path_to_folder):
		os.mkdir(path_to_folder)

def convert_to_bytes(size):
	"""
	Will convert the size in any disk space format to bytes.

	:param size: The size of the database in any units
	:type  size: str.

	returns data_size: The size of the database in bytes
	rtype   data_size: float
	"""
	data_size = float(size[:-2]); data_size_type = size[-2:]
	if data_size_type == 'KB':
		data_size *= (1024.0 ** 1)
	elif data_size_type == 'MB':
		data_size *= (1024.0 ** 2)
	elif data_size_type == 'GB':
		data_size *= (1024.0 ** 3)
	elif data_size_type == 'TB':
		data_size *= (1024.0 ** 4)
	else:
		exit('Error')
	return data_size

class GA_Recording_System:
	"""
	This class is designed to record the clsuters that are created during the Organisms program run.

	:param ga_recording_information: This is a dictionary that contains all the information that it needs to record clusters made during the Organisms program as the user desires.
	:type  ga_recording_information (dict.):
	"""
	def __init__(self,ga_recording_information):
		####################################################################################################
		# determine the type of ga_recording_scheme to use, and the other settings associated to each ga_recording_scheme.
		if ga_recording_information == None or ga_recording_information == {}:
			self.ga_recording_information = {'ga_recording_scheme': 'None'}
		else:
			self.ga_recording_information = ga_recording_information
		if self.ga_recording_information['ga_recording_scheme'] in ['None', 'none', None]:
			self.ga_recording_information = {'ga_recording_scheme': 'None'}
		self.ga_recording_scheme = self._get_parameter('ga_recording_scheme',if_not_given_response='None')
		# Determine the name of this recording instance. THis is the name of the folder clusters will be recorded into.
		self.ga_recording_system_name = 'Recorded_Data'
		self.path_to_write_to = os.getcwd()+'/'+self.ga_recording_system_name
		self.ga_recording_system_name = self._get_parameter('ga_recording_system_name',if_not_given_response=self.ga_recording_system_name)
		# determine when to collect data during generations
		self.exclude_recording_cluster_screened_by_diversity_scheme = self._get_parameter('exclude_recording_cluster_screened_by_diversity_scheme',if_not_given_response=True)
		self.limit_number_of_clusters_recorded = self._get_parameter('limit_number_of_clusters_recorded',if_not_given_response='None') # determine the maximum number of clusters to record
		self.limit_size_of_database = self._get_parameter('limit_size_of_database',if_not_given_response='None')
		if not self.limit_size_of_database == None:
			self.limit_size_of_database = convert_to_bytes(self.limit_size_of_database)
		self.show_GA_Recording_Database_check_percentage = self._get_parameter('show_GA_Recording_Database_check_percentage',if_not_given_response=False)
		####################################################################################################
		# Set setting depending on the ga_recording_scheme chosen to use
		if self.ga_recording_scheme == 'All':
			self.clusters_database = GA_Recording_Database(self.path_to_write_to,self.limit_number_of_clusters_recorded,self.limit_size_of_database,ga_recording_scheme=self.ga_recording_scheme,show_GA_Recording_Database_check_percentage=self.show_GA_Recording_Database_check_percentage)
			self.directory = os.getcwd()+'/'+self.ga_recording_system_name
		elif self.ga_recording_scheme == 'Limit_energy_height':
			self.limit_energy_height_of_clusters_recorded = self._get_parameter('limit_energy_height_of_clusters_recorded',if_not_given_response=float('inf')) # Determine the maximum energy of a cluster to record, above the energy of the lowest energetic clusters obtained from this genetic algorithm.
			if self.limit_energy_height_of_clusters_recorded == float('inf'):
				print('Something to note in GA_Recording_System, in GA_Recording_System.py')
				print("You have set ga_recording_scheme == 'Limit_energy_height', but you have set (either written in you Run.py script or given by default) your limit_energy_height_of_clusters_recorded to greatest value it could be")
				print("It would be best to change your ga_recording_scheme input in your ga_recording_information dictionary to :")
				print("ga_recording_scheme == 'All'")
				print("Manually change this in your Run.py script before continuing.")
				exit('The genetic algorithm will finish without completing.')			
			self.clusters_database = GA_Recording_Database(self.path_to_write_to,self.limit_number_of_clusters_recorded,self.limit_size_of_database,ga_recording_scheme=self.ga_recording_scheme,limit_energy_height_of_clusters_recorded=self.limit_energy_height_of_clusters_recorded,show_GA_Recording_Database_check_percentage=self.show_GA_Recording_Database_check_percentage)
		elif self.ga_recording_scheme == 'Set_higher_limit':
			self.upper_energy_limit = self._get_parameter('upper_energy_limit',if_not_given_response=float('inf'))
			if self.upper_energy_limit == float('inf'):
				print('Something to note in GA_Recording_System, in GA_Recording_System.py')
				print("You have set ga_recording_scheme == 'Limit_energy_height', but you have set (either written in you Run.py script or given by default) your upper_energy_limit to greatest value it could be")
				print("It would be best to change your ga_recording_scheme input in your ga_recording_information dictionary to :")
				print("ga_recording_scheme == 'All'")
				print("Manually change this in your Run.py script before continuing.")
				exit('The genetic algorithm will finish without completing.')	
			self.clusters_database = GA_Recording_Database(self.path_to_write_to,self.limit_number_of_clusters_recorded,self.limit_size_of_database,ga_recording_scheme=self.ga_recording_scheme,upper_energy_limit=self.upper_energy_limit,show_GA_Recording_Database_check_percentage=self.show_GA_Recording_Database_check_percentage)
		elif self.ga_recording_scheme == 'Set_energy_limits':
			self.lower_energy_limit = self._get_parameter('lower_energy_limit',if_not_given_response=-float('inf'))
			self.upper_energy_limit = self._get_parameter('upper_energy_limit',if_not_given_response=float('inf'))
			if (self.lower_energy_limit == -float('inf')) and (self.upper_energy_limit == float('inf')):
				print('Something to note in GA_Recording_System, in GA_Recording_System.py')
				print("You have set ga_recording_scheme == 'Set_energy_limits', but you have set (either written in you Run.py script or given by default) your upper energy limit to the greatest value it could be and the lower energy limit to the lowest value it could be.")
				print("It would be best to change your ga_recording_scheme input in your ga_recording_information dictionary to :")
				print("ga_recording_scheme == 'All'")
				print("Manually change this in your Run.py script before continuing.")
				exit('The genetic algorithm will finish without completing.')
			self.clusters_database = GA_Recording_Database(self.path_to_write_to,self.limit_number_of_clusters_recorded,self.limit_size_of_database,ga_recording_scheme=self.ga_recording_scheme,lower_energy_limit=self.lower_energy_limit,upper_energy_limit=self.upper_energy_limit,show_GA_Recording_Database_check_percentage=self.show_GA_Recording_Database_check_percentage)
		elif self.ga_recording_scheme == 'None':
			pass
		else:
			exit('Error')
		####################################################################################################
		# make recording system folder
		if not self.ga_recording_scheme == 'None':
			self.directory = os.getcwd()+'/'+self.ga_recording_system_name
			make_folder(self.directory)
		####################################################################################################
		# Get the name for the information file.
		self.RC_Information_TXT_name = self.ga_recording_system_name+'_Information.txt'
		# Other features of the genetic algorithm to monitor.
		self.record_initial_population = self._get_parameter('record_initial_population',if_not_given_response=False)
		####################################################################################################
		# create required variables and files that are needed to record the stages of the algorithm after the desired generations as specified in saving_points_of_GA
		self.saving_points_of_GA = self._get_parameter('saving_points_of_GA',if_not_given_response=[])
		if not self.saving_points_of_GA == []:
			self.saving_points_in_GA_name = 'Saved_Points_In_GA_Run'
			self.directory_for_saving_points_in_GA = os.getcwd()+'/'+self.saving_points_in_GA_name
			if not os.path.exists(self.directory_for_saving_points_in_GA):
				os.mkdir(self.directory_for_saving_points_in_GA)
		####################################################################################################

	def __repr__(self):
		return str(self.__dict__)

	def add_metadata(self):
		"""
		This method is designed to assign the metadata o the ASE database, as in some versions of ASE this can not happen until at least one cluster has been added to the ASE database.
		"""
		if 'clusters_database' in self.__dict__.keys():
			self.clusters_database.add_metadata()

	def _get_parameter(self,parameter,if_not_given_response=None):
		"""
		This definition is designed to assign self variables to instance of Recording_Clusters. These variables are
		the variables required by Recording_Clusters

		This has been designed to be a private definition. 
		
		:param parameter: This is the parameter to obtain from self.ga_recording_information
		:type  parameter: str.
		:param if_not_given_response: This is the default value to assign the parameter parameter to in this class if it hasn't been given by the user. 
		:type  if_not_given_response: Any

		returns the value to assign a certain parameter to in this class.
		rtype   Any

		"""
		for key, value in self.ga_recording_information.items():
			if key == parameter:
				if value == 'inf':
					value = float('inf') 
				return value
		if if_not_given_response == None:
			print('Error in class Recording_Clusters, in Recording_Clusters.py')
			print('The variable "'+str(parameter)+'" was not entered into Recording_Clusters')
			print('Check this')
			import pdb; pdb.set_trace()
			exit()
		elif if_not_given_response == 'None':
			return None
		else:
			return if_not_given_response

	def add_collection(self,collection,offspring_to_remove):
		"""
		This will add the clusters to the GA_Recording_System database

		:param collection: The collection to be recorded
		:type  collection: Organisms.GA.Collection
		:param offspring_to_remove: This is a list of all the clusters not to write to the database, as they have been removed by the diversity operator, depending on if self.exclude_recording_cluster_screened_by_diversity_scheme is True or False.
		:type  offspring_to_remove: list of int
		"""
		if self.exclude_recording_cluster_screened_by_diversity_scheme:
			offspring_to_remove = offspring_to_remove
		else:
			offspring_to_remove = []
		if self.ga_recording_scheme in ['All','Limit_energy_height','Set_higher_limit','Set_energy_limits']:
			self.clusters_database.add_collection_to_database(collection, offspring_to_remove)

	def update_cluster_in_database_for_if_in_population(self,clusters_in_the_population,energies_of_clusters_removed_from_the_population):
		"""
		This method will update clusters in the database if that cluster was ever accepted into the population. 

		:param clusters_in_the_population: This is a list of the names of clusters that are in the population.
		:type  clusters_in_the_population: list of int
		:param energies_of_clusters_removed_from_the_population: This is a list of the energies of clusters that are in the population.
		:type  energies_of_clusters_removed_from_the_population: list of int

		"""
		if self.ga_recording_scheme == 'None':
			return
		self.clusters_database.update_cluster_in_database_for_if_in_population(clusters_in_the_population, energies_of_clusters_removed_from_the_population)

	#-----------------------------------------------------------------------------------------------------------------
	#-----------------------------------------------------------------------------------------------------------------
	#-----------------------------------------------------------------------------------------------------------------
	# Methods for restoring the ga_recording_system from some generation. 

	def resume_ga_recording_system_from_current_generation(self,resume_from_generation):
		"""
		Will check and restore the ga_recording_system is restored for a generation. 

		:param resume_from_generations: The generation that your genetic algorithm is up to.
		:type  resume_from_generation: int

		"""
		if self.ga_recording_scheme == 'None':
			return
		self.check_clusters_in_database(resume_from_generation)
		self.import_information_from_database(resume_from_generation)

	def check_clusters_in_database(self,generation):
		"""
		Will check the clusters in the database and remove any cluster that was created during a most recent unsuccessful generation.

		:param generation: The current generation
		:type  generation: int

		"""
		if self.ga_recording_scheme == 'None':
			return
		self.clusters_database.check_clusters_in_database(generation)

	def import_information_from_database(self,current_generation):
		"""
		Import any data from the database if it is needed. This should only be needed for the 'Limit_energy_height' scheme.

		:param current_generation: The current generation that your genetic algorithm trial is being resumed from
		:type  current_generation: int

		"""
		if self.ga_recording_scheme == 'None':
			return
		if self.ga_recording_scheme == 'All':
			return
		self.clusters_database.import_information_from_database(current_generation)

	#-----------------------------------------------------------------------------------------------------------------
	#-----------------------------------------------------------------------------------------------------------------
	#-----------------------------------------------------------------------------------------------------------------
	# The follow is methods for recording the initial population or the population at different generations. 

	def record_initial_populations(self,population):
		"""
		Will record the clusters in the initial population into GA_Recording_System.

		:param population: The initial population
		:type  population: Organisms.GA.Population
		"""
		if self.record_initial_population:
			name_of_database = 'Initial_'+population.name
			path_to_write_to = os.getcwd()+'/'+'Initial_'+population.name
			self.record_collection(population,name_of_database,path_to_write_to=path_to_write_to)

	def record_population_at_generation(self,population,current_generation):
		"""
		Will record the clusters in the initial population into GA_Recording_System at the current generation.

		:param population: The current population after the generation has completed
		:type  population: Organisms.GA.Population
		:param current_generation: The generation that your genetic algorithm is up to.
		:type  current_generation: int

		"""
		if current_generation in self.saving_points_of_GA:
			name_of_database = str(population.name)+'_Gen_'+str(current_generation)
			path_to_write_to = self.directory_for_saving_points_in_GA+'/'+str(population.name)+'_Gen_'+str(current_generation)
			self.record_collection(population,name_of_database,path_to_write_to=path_to_write_to)

	def record_collection(self,collection,name_of_database,path_to_write_to):
		"""
		This records an identical copy of the clusters in the population at a certain generation. 

		:param collection: The collection to record into GA_Recording_System.
		:type  collection: Organisms.GA.Collection
		:param name_of_database: Name of the database oto connect to
		:type  name_of_database: str.
		:param path_to_write_to: Path of the database or folder of xyz files to write to.
		:type  path_to_write_to: str. 
			
		"""
		os.mkdir(path_to_write_to)
		database = connect(path_to_write_to+'/'+name_of_database+'.db')
		for cluster in collection:
			database.write(cluster.deepcopy(),name=cluster.name,gen_made=cluster.gen_made,cluster_energy=cluster.energy)
		database.metadata = {'title': str(name_of_database), 'key_descriptions': {'systems': ('Cluster', 'The cluster', ''), 'gen_made': ('Generation Created', 'The generation tthe cluster was created.', ''), 'cluster_energy': ('Cluster Energy', 'Potential energy of the cluster.', 'eV'), 'name': ('Name', 'Name of the cluster.', '')}, 'default_columns': ['name', 'cluster_energy', 'id']}
	


"""

MatingProcedure.py, 13/04/2017, Geoffrey R. Weal

This program is designed to run the mating proceedure of the Genetic Algorithm.

"""

from ase import Atom, Atoms
import random as rand
import numpy as np
import copy
from Organisms.GA.ExternalDefinitions import get_elemental_makeup, AtomInClusterPosition, InclusionRadiusOfCluster
from Organisms.GA.Cluster import Cluster
from Organisms.GA.Population import Population
from collections import Counter

#population_type_example = Population('example',-1,False,None,False)
#cluster_type_example = Cluster()

class Crossover:
	"""
	This class is designed to perform the mating proceedure of the genetic algorithm. This will produce a cluster that is an outcome of the mating proceedure.

	:param crossover_type: This is the type of mating proceedure the user would like to use. There are currently a few options implimented into this mating proceedure:

		* "CAS_weighted"-: Cut and Splice - Deavon and Ho - weighted by fitness of parents (CAS_weighted)
		* "CAS_random":    Cut and Splice - Deavon and Ho - cut a random percent x% of parent 1 and (100-x)% of parent 2 (CAS_random)
		* "CAS_half":      Cut and Splice - Deavon and Ho - cut both parents by half (CAS_half)
		* "CAS_custom_XX":   Cut and Splice - Deavon and Ho - cut a random percent XX% of parent 1 and (100-XXX)% of parent 2. To use this set crossType = CAS_custom_XX, where XX is a float of your choice between 0 and 100.
	:type  crossover_type: str.
	:param r_ij: the maximum bond distance between atoms in the cluster. This should be the largest value possible for your cluster.
	:type  r_ij: float
	:param vacuumAdd: The vacuum around the cluster
	:type  vacuumAdd: float
	:param size_of_clusters: The number of atoms in the cluster.
	:type  size_of_clusters: int

	"""
	def __init__(self,crossover_type,r_ij,vacuumAdd,size_of_clusters):
		self.crossover_type = crossover_type
		self.r_ij = r_ij
		self.vacuumAdd = vacuumAdd
		self.size_of_clusters = size_of_clusters
		# -------------------------------------------- #
		# Assign a method to use for the mating proceedure.
		if self.crossover_type == "CAS_weighted":
			self.mating_method = self.Cut_and_Splice_Devon_and_Ho
		elif self.crossover_type == "CAS_random":
			self.half_index = self.half_index_random_method()
			self.mating_method = self.Cut_and_Splice_Devon_and_Ho
		elif self.crossover_type == "CAS_half":
			self.half_index = self.half_index_half_method()
			self.mating_method = self.Cut_and_Splice_Devon_and_Ho
		elif "CAS_custom_" in self.crossover_type:
			self.half_index = self.half_index_custom_method()
			self.mating_method = self.Cut_and_Splice_Devon_and_Ho
		else:
			print("Error in def mating in class MatingProcedure, in MatingProcedure.py: No Mating method selected")
			print("Mating Proceedures available:")
			print("\tCut and Splice - Deavon and Ho - weighted by fitness of parents (CAS_weighted)")
			print("\tCut and Splice - Deavon and Ho - cut a random percent x% of parent 1 and (100-x)% of parent 2 (CAS_random)")
			print("\tCut and Splice - Deavon and Ho - cut both parents by half (CAS_half)")
			print("\tCut and Splice - Deavon and Ho - cut a random percent XXX% of parent 1 and (100-XXX)% of parent 2 (CAS_custom_XXX). To use this set crossType = CAS_custom_XXX, where XXX is a float of your choice between 0 and 100.")
			exit('This program will finish without completing')
		# -------------------------------------------- #

	def run(self,run_input):
		"""
		Run this the Mating Proceedure and gives an offspring.

		:param population: This is the population to choose clusters from to mate together. 
		:type  population: Organisms.GA.Population

		:returns: offspring
		:rtypes: Organisms.GA.Cluster

		"""
		parents = self.pickParentsFromThePopulation(run_input)
		#parents = self.centre_parents_about_origin(parents) # put a 
		offspring = self.mating(parents)
		#self.centre_offspring_at_centre_of_cell(offspring)
		return offspring

	def pickParentsFromThePopulation(self,population):
		"""
		This definition will pick the parents for mating.

		:param population: This is the population to choose clusters from to mate together. 
		:type  population: Organisms.GA.Population

		:returns: two clusters from the population that will be mated together to give the offspring.
		:rtypes: (Organisms.GA.Cluster, Organisms.GA.Cluster)

		"""
		pairPos = self.roulette(population)
		parent1 = population[pairPos[0]].deepcopy()
		parent2 = population[pairPos[1]].deepcopy()
		return [parent1, parent2]

	def roulette(self,population):
		"""
		Performed the roulette wheel method to obtain the parents for a Mating proceedure

		Reference: https://pubs.rsc.org/en/content/articlepdf/2003/dt/b305686d

		:param population: This is the population to choose clusters from to mate together. 
		:type  population: Organisms.GA.Population

		:returns: Will return a tuple of two integers that represent the indices of clusters in the population to use as parent to mate together to give a new offspring.
		:rtypes: [int, int]

		"""
		index_pair = []; index_pair_names = []; population_size = len(population)
		# Pick parents
		while len(index_pair) < 2:
			chosen_cluster_index = rand.randrange(0,population_size,1) # Pick a random cluster to assess.
			chosen_cluster = population[chosen_cluster_index]
			if not chosen_cluster.name in index_pair_names:
				randomFit = rand.uniform(0,1) # Choose a random fitness criteria value.
				# If the random cluster's fitness value is higher than the fitness criteria 
				# value (and is not already a parent), accept the cluster as a parent. 
				if chosen_cluster.fitness > randomFit:
					index_pair.append(chosen_cluster_index)
					index_pair_names.append(chosen_cluster.name)
		# This is a method to check that all is well with the roulette method. Turn on if you are ever playing around with it
		if len(index_pair_names) != len(set(index_pair_names)) or len(index_pair) != len(set(index_pair)):
			print("Error in def roulette in class MatingProcedure, in MatingProcedure.py")
			print("A cluster has been selected at least twice by the roulette wheel method.")
			if len(index_pair_names) != len(set(index_pair_names)):
				print('Two of the clusters have the same dir (i.e. this means they are the same cluster).')
			if len(index_pair) != len(set(index_pair)):
				print('Two of the clusters have the same index in the population.')
			print('Clusters that have been selected: '+str(index_pair_names))
			print('The indices of the clusters in the population that have been selected: '+str(index_pair))
			print('Check this.')
			import pdb; pdb.set_trace()
			exit('This program will finish without completing')
		return index_pair

	def tournament(self,population):
		"""
		Performed the tournament method to obtain the parents for a Mating proceedure

		Reference: https://pubs.rsc.org/en/content/articlepdf/2003/dt/b305686d

		:param population: This is the population to choose clusters from to mate together. 
		:type  population: Organisms.GA.Population
		
		:returns: Will return a tuple of two integers that represent the indices of clusters in the population to use as parent to mate together to give a new offspring.
		:rtypes: [int, int]
		"""
		number_of_clusters_in_tournament = X
		clusters_in_tournament = []; population_size = len(population)
		clusters_in_tournament_names = []
		# Pick parents
		while len(clusters_in_tournament) < number_of_clusters_in_tournament:
			chosen_cluster_index = rand.randrange(0,population_size,1) # Pick a random cluster to assess.
			chosen_cluster = population[chosen_cluster_index]
			if not chosen_cluster.name not in clusters_in_tournament_names:
				clusters_in_tournament_names.append(chosen_cluster.name)
				clusters_in_tournament.append((chosen_cluster_index,chosen_cluster.fitness))
		if len(clusters_in_tournament) != len(set(clusters_in_tournament)):
			print("Error in def tournament in class MatingProcedure, in MatingProcedure.py")
			print("A cluster has been selected at least twice by the tournament method.")
			print('Clusters that have been selected (index in pop, cluster name): '+str(clusters_in_tournament))
			print('Check this.')
			import pdb; pdb.set_trace()
			exit('This program will finish without completing')
		clusters_in_tournament.sort(key = lambda x: x[1], reverse=True)
		index_pair = (clusters_in_tournament[0][0], clusters_in_tournament[1][0])
		return index_pair

	def centre_parents_about_origin(self,parents):
		"""
		This method will center the parents about the zero point. This is to make sure that their two halfs will allign correctly when they are cut and spliced to form a new offsping

		:param parents: This is the population to choose clusters from to mate together. 
		:type  parents: [Organisms.GA.Cluster, Organisms.GA.Cluster]
		
		:returns: The parents which have been centered about the (0,0,0) point of the unit cell
		:rtypes: [Organisms.GA.Cluster, Organisms.GA.Cluster]

		"""
		for parent in parents:
			parent.center(vacuum=None, axis=(0, 1, 2), about=0.)
		return parents

	def centre_offspring_at_centre_of_cell(self,offspring):
		"""
		This method will center the offspring as required. 

		:param offspring: The offspring cluster. 
		:type  offspring: Organisms.GA.Cluster

		"""
		offspring.centre_cluster_at_centre_of_cell(self.vacuumAdd)

	def mating(self, parents):
		"""
		This definition is designed as a switch to choose the mating method the user wishes to use for the genetic algorithm.

		:param parents: This is the population to choose clusters from to mate together. 
		:type  parents: [Organisms.GA.Cluster, Organisms.GA.Cluster]

		:returns: Returns an offspring that has been created from mating two parent clusters together. 
		:rtypes:  Organisms.GA.Cluster

		"""
		return self.mating_method(parents) # returns a new offspring

	#################################################################################
	#################################################################################

	def Cut_and_Splice_Devon_and_Ho(self,parents):
		"""
		This definition is designed to perform the Cut and Splice Method as specified by 
		Devon and Ho to mate the parents to give the offspring. The method works as follows:
		
			1. Rotate each parent by some random amount in the theta and phi directions.
			2. Sort the assignment of atoms in the cluster from most positive to most negative z value.
			3. Perform a Cut and Splice proceedure, where the parents are cut in some way and spliced together to give the offspring. 

		:param parents: This is the population to choose clusters from to mate together. 
		:type  parents: [Organisms.GA.Cluster, Organisms.GA.Cluster]
		
		:returns: Returns an offspring that has been created from mating two parent clusters together. 
		:rtypes:  Organisms.GA.Cluster
		"""
		for i in range(len(parents)):
			self.rotate(parents[i]) # Rotate parent parents[1] = self.rotate(parents[1]) 
			parents[i].sortZ() # sort atoms in parent from most positive to most negative value of z
		offspring = self.mate_Cut_and_Splice(parents) # Perform the Cut and Splice Method
		return offspring

	def rotate(self,cluster):
		"""
		Rotate the cluster by some random angle in the theta and phi directions.
		
		:param cluster: The cluster to be randomly rotated
		:type  cluster: Organisms.GA.Cluster

		"""
		#randomly select the value of theta and phi. Change to degrees since change in ase 3.14.0
		# see https://mathworld.wolfram.com/EulerAngles.html for how the angles work.
		phi = rand.uniform(0,360.0)
		theta = rand.uniform(0,180.0) # angle theta in [0,pi]
		psi = rand.uniform(0,360.0)
		# Create the matrix representation to rotate the cluster in the following way:
		#	Look at Euler Angles about this: http://mathworld.wolfram.com/EulerAngles.html
		cluster.euler_rotate(phi=phi, theta=theta, psi=psi, center=(0, 0, 0)) # changed to allign with format from ase >= 3.14.0
		#return cluster

	def mate_Cut_and_Splice(self,parents):
		"""
		This method will perform the mating proceedure as specified by the user.
		This can been designed to be used with multiple parents, however currently it is set up for 2 parents.
		It has also been developed for mating multi-metallic (multi-elemental) clusters together.

		:param parents: This is the population to choose clusters from to mate together. 
		:type  parents: [Organisms.GA.Cluster, Organisms.GA.Cluster]
		
		:returns: Returns an offspring that has been created from mating two parent clusters together. 
		:rtypes:  Organisms.GA.Cluster	

		"""
		#elemental_makeup_of_parents = self.mate_Cut_and_Splice_error_checking_1(parents) # Error Checking. Uncomment this if you are ever editing the crossover method
		# if "weighed" is chosen, use the fitness values of parents to find the cut points in the clusters.
		if self.crossover_type == "CAS_weighted":
			self.half_index = self.half_index_weighted_method(parents)
		if self.crossover_type == "CAS_random":
			self.half_index = self.half_index_random_method()
		# Get half of first parent
		offspring = parents[0][:self.half_index:]
		elementsInParent1 = parents[1].get_elemental_makeup() #elemental_makeup_of_parents[1]
		if len(elementsInParent1.keys()) > 1:
			# Count the difference in elements between the parent and the current cluster. This is so the
			# offspring has the same numbers of elements as its parents
			elements_in_current_offspring = offspring.get_elemental_makeup()
			elements_to_add = elementsInParent1 - elements_in_current_offspring
			for atom_to_add in parents[1][::-1]:
				if elements_to_add[atom_to_add.symbol] > 0 and not AtomInClusterPosition(atom_to_add,offspring):
					offspring.append(atom_to_add)
					elements_to_add[atom_to_add.symbol] -= 1
					if not any(not value == 0 for value in elements_to_add.values()):
						break
		else:
			#for atom_to_add in parents[1][self.half_index::]:
			#	if not AtomInClusterPosition(atom_to_add,offspring):
			#		offspring.append(atom_to_add)
			offspring += parents[1][self.half_index::]
		#self.mate_Cut_and_Splice_error_checking_2(elemental_makeup_of_parents,offspring) # Error Checking. Uncomment this if you are ever editing the crossover method
		return offspring

	def mate_Cut_and_Splice_error_checking_1(self,parents):
		"""
		This is a method for checking if there are any issues with the Cut_and_splice method. 

		This method is only needed when you are developing or modifying a mating proceedure, to help with debugging. Otherwise, you dont need to use this algorithm

		:param parents: This is the population to choose clusters from to mate together. 
		:type  parents: [GA.Cluster, GA.Cluster]

		:returns: A list which indicates the types of elements, and the number of those elements, in the cluster
		:rtype: {str: int, ...} 
		"""
		elemental_makeup_of_parents = []
		for parent in parents:
			elemental_makeup_of_parents.append(parent.get_elemental_makeup())
		# check that the parents have the same elements and amount of every element:
		for index in range(1,len(elemental_makeup_of_parents)):
			if not elemental_makeup_of_parents[index] == elemental_makeup_of_parents[0]:
				print('Error in def mate_Cut_and_Splice: The parents do not have the same element count.')
				print('Element Count of Parents')
				for index in range(len(elemental_makeup_of_parents)):
					print('Parent ' + str(index+1) + ': ' + str(elemental_makeup_of_parents[index]))
				print('Check this')
				import pdb; pdb.set_trace()
				exit('This program will finish without completing')
		return elemental_makeup_of_parents


	def mate_Cut_and_Splice_error_checking_2(self,elemental_makeup_of_parents,offspring):
		"""
		This is a method for checking if there are any issues with the Cut_and_splice method. 

		This method is only needed when you are developing or modifying a mating proceedure, to help with debugging. Otherwise, you dont need to use this algorithm

		:param elemental_makeup_of_parents: A list which indicates the types of elements, and the number of those elements, in the cluster
		:type  elemental_makeup_of_parents: {str: int, ...} 
		:param offspring: The offspring cluster. 
		:type  offspring: Organisms.GA.Cluster

		"""
		elemental_makeup_of_offspring = get_elemental_makeup(offspring)
		if not elemental_makeup_of_offspring == elemental_makeup_of_parents[0]:
			print('Error in def mate_Cut_and_Splice: The parents do not have the same element count.')
			print("ERROR: Parent and offspring elemental makeup are not the same: No of Atoms in;")
			print('-------------------------')
			print('Element Count of Parents')
			for index in range(len(elemental_makeup_of_parents)):
				print('Parent ' + str(index+1) + ': ' + str(elemental_makeup_of_parents[index]))
			print('-------------------------')
			print('Element Count of Offspring')
			print('Offspring: ' + str(offspring.get_elemental_makeup()))
			print('-------------------------')
			import pdb; pdb.set_trace()
			exit('This program will finish without completing')


	def half_index_weighted_method(self,parents):
		"""
		This method will determine how to cut the cluster based on the atom to divide from, where the atoms have been numbered in order of the z axis.

		This version of the method divides the parents based on their relative fitnesses.

		:param parents: This is the population to choose clusters from to mate together. 
		:type  parents: [Organisms.GA.Cluster, Organisms.GA.Cluster]

		:returns: The atom number in the cluster to cut the atom.
		:rtype: int
		"""
		fitPair = []
		for parent in parents:
			fitPair.append(parent.fitness)
		fit1 = fitPair[0]; fit2 = fitPair[1]
		return int(self.size_of_clusters*(fit1/(fit1+fit2))) # This is half_index

	def half_index_random_method(self):
		"""
		This method will determine how to cut the cluster based on the atom to divide from, where the atoms have been numbered in order of the z axis.

		This version of the method divides the parents at some random atom indice

		:returns: The atom number in the cluster to cut the atom.
		:rtype: int
		"""
		half_index = int(self.size_of_clusters*rand.random())
		return half_index

	def half_index_half_method(self):
		"""
		This method will determine how to cut the cluster based on the atom to divide from, where the atoms have been numbered in order of the z axis.

		This version of the method divides the parents in half.

		:returns: The atom number in the cluster to cut the atom.
		:rtype: int
		"""
		half_index = int(self.size_of_clusters*(1.0/2.0))
		return half_index

	def half_index_custom_method(self,cross_type):
		"""
		This method will determine how to cut the cluster based on the atom to divide from, where the atoms have been numbered in order of the z axis.

		This version of the method divides the parents based on a percentage of atoms to take from parent 1

		Here, percentage of parent1 is taken, while 1.0-percentage of parent2 two is taken.

		:param crossover_type: This is the type of mating proceedure the user would like to use. This should be set to "CAS_custom_XX", where XX is the percentage of parent 1 to be cut.
		:type  crossover_type: str.

		:returns: The atom number in the cluster to cut the atom.
		:rtype: int
		"""
		percentage = cross_type.replace('custom_','')
		try:
			percentage = float(percentage)
		except ValueError:
			exit("Error in MatingProcedure.py using Cut and Splice - Deavon and Ho - Custom. " + str(percentage) + " is not a float.") 
		if percentage < 0.0 or percentage > 100:
			print("Error in MatingProcedure.py using Cut and Splice - Deavon and Ho - Custom.")
			print("Custom percentage must be between 0 and 100. Percentage = " + str(percentage))
			exit('This program will finish without completing')
		half_index = int(self.size_of_clusters*(percentage/100.0))
		return half_index





import numpy as np

def get_SCM_methods(SCM_Scheme):
	"""
	This method will provide the methods needed to obtain the CNA_profile and similarity_profile, using either the T-SCM or the A-SCM. 

	:param SCM_Scheme: This determines if the SCM scheme being used is the "T-SCM" or "A-SCM". Must be one of those two schemes.
	:type  SCM_Scheme: str.
	"""
	if SCM_Scheme == 'A-SCM':
		from Organisms.GA.SCM_Scripts.AC_SCM_Methods import get_CNA_profile, get_CNA_similarities
	elif SCM_Scheme == 'T-SCM':
		from Organisms.GA.SCM_Scripts.TC_SCM_Methods import get_CNA_profile, get_CNA_similarities
	else:
		print('Error in initialisation of CNA_Database, in CNA_Database.py')
		print('self.SCM_Scheme must be either "T-SCM" or "A-SCM".')
		print('self.SCM_Scheme = '+str(SCM_Scheme))
		print('Check this.')
		import pdb; pdb.set_trace()
		exit()
	return get_CNA_profile, get_CNA_similarities

def get_rCut_values(rCut_low,rCut_high,rCut_resolution):
	"""
	This method will give a lost of rCut values that you want to perform the CNA across, in Angstroms

	:parm rCut_low: This is the lowest rCut value you want to obtain the CNA for.
	:type rCut_low: float
	:parm rCut_high: This is the highest rCut value you want to obtain the CNA for.
	:type rCut_high: float
	:parm rCut_resolution: This is the resolution of the rCut values you are performing all the CNA's with. The resolution can be given as one of two forms. 

		* If given as a float, rCut_resolution is the difference between rCut values that will be sampled. Note that your input for rCut_resolution is not divisible by (rCut_high-rCut_low), then rCut_high will not be included in rCuts.
		* If given as a int, rCut_resolution is the number of rCut values in the list rCuts, evenly distributed between (and including) rCut_low and rCut_high.

	:type rCut_resolution: float or int

	:returns: A list of all the rCut values you want the CNA to sample.
	:rtype:   list of floats
	"""
	if isinstance(rCut_resolution,float):
		rCuts = list(np.arange(rCut_low,rCut_high+rCut_resolution,rCut_resolution))
	elif isinstance(rCut_resolution,int):
		rCuts = list(np.linspace(rCut_low,rCut_high,num=rCut_resolution,endpoint=True))
	else:
		print('Error in class CNA_Diversity_Scheme, in CNA_Diversity_Scheme.py')
		print('self.rCut_resolution must be either a float or an int.')
		print('self.rCut_resolution = '+str(rCut_resolution))
		print('Check this.')
		print('This program will not end.')
		exit()
	for index in range(len(rCuts)-1,-1,-1):
		rCuts[index] = round(rCuts[index],10)
	return rCuts

def get_rCuts(self,Predation_Information):
	"""
	Obtain the values for values of rCut the user wishes to investigate.

	:param Predation_Information: This contains all the information needed by the Predation Operator you want to use to run.
	:type  Predation_Information: dict.

	:returns: A list of all the rCut values you want the CNA to sample.
	:rtype:   list of floats
	
	"""
	# An initial check to make sure of you have parameters
	if ('lattice_constant' in Predation_Information):
		if any([(name in ['rCut','rCut_low','rCut_high','rCut_resolution']) for name in Predation_Information]):
			print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
			print('You have included the "lattice_constant" parameter in your Predation_Information')
			print("but you have also included 'rCut' and/or 'rCut_low','rCut_high','rCut_resolution' in the Predation_Information")
			print('Check the settings in the Predation_Information')
			print('Predation_Information = '+str(Predation_Information))
			import pdb; pdb.set_trace()
			exit('This program will exit without completing')
	else:
		if any([(name in ['single_nn_measurement','nn_low','nn_high','nn_resolution']) for name in Predation_Information]):
			print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
			print('You have not included the "lattice_constant" parameter in your Predation_Information')
			print("but you have included 'single_nn_measurement' and/or 'nn_low','nn_high','nn_resolution' in the Predation_Information")
			print('Check the settings in the Predation_Information')
			print('Predation_Information = '+str(Predation_Information))
			import pdb; pdb.set_trace()
			exit('This program will exit without completing')

	# Perform a check on the single_nn_measurement, nn_low, nn_high, and nn_resolution inputs in Predation_Information
	if ('lattice_constant' in Predation_Information):
		self.lattice_constant = float(Predation_Information['lattice_constant'])
		self.fnn_distance = lattice_constant / (2.0 ** 0.5)
		if (('nn_low' in Predation_Information) or ('nn_high' in Predation_Information) or ('nn_resolution' in Predation_Information)):
			if not (('nn_low' in Predation_Information) and ('nn_high' in Predation_Information) and ('nn_resolution' in Predation_Information)):
				print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
				print('If you are wanting to have a range of nn values, you need to include the following in Predation_Information:')
				print('\t* nn_low')
				print('\t* nn_high')
				print('\t* nn_resolution')
				print('Make sure you have all of these in Predation_Information.')
				print('Predation_Information = '+str(Predation_Information))
				import pdb; pdb.set_trace()
				exit('This program will exit without completing')
			elif ('single_nn_measurement' in Predation_Information):
				print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
				print('In Predation_Information, you have included single_nn_measurement as well as nn_low, nn_high, and/or nn_resolution.')
				print('You can only enter into Predation_Information either single_nn_measurement, or all of nn_low, nn_high, and nn_resolution')
				print('Only inlclude the following if you want to have a range of nn values:')
				print('\t* nn_low')
				print('\t* n_high')
				print('\t* rn_resolution')
				print('Only include single_nn_measurement if you want to include only want to sample the similarity at one value of single_nn_measurement.')
				print('Predation_Information = '+str(Predation_Information))
				import pdb; pdb.set_trace()
				exit('This program will exit without completing')
			else:
				multiple_rCut = True; simple_rCut = False
				self.nn_low = float(Predation_Information['nn_low'])
				if not (1.0 <= nn_low <= 2.0):
					print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
					print('nn_low must be between 1.0 and 2.0')
					print('nn_low = '+str(nn_low))
					print('Check this')
					exit('This program will exit without completing')
				self.nn_high = float(Predation_Information['nn_high'])
				if not (1.0 <= nn_high <= 2.0):
					print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
					print('nn_high must be between 1.0 and 2.0')
					print('nn_high = '+str(nn_high))
					print('Check this')
					exit('This program will exit without completing')
				self.nn_resolution = Predation_Information['nn_resolution']
				if not nn_resolution % 1.0 == 0:
					print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
					print('nn_resolution must be a whole number')
					print('nn_resolution = '+str(nn_resolution))
					print('Check this')
					exit('This program will exit without completing')
				self.nn_resolution = int(nn_resolution)
				self.rCut_low = self.fnn_distance *self. nn_low
				self.rCut_high = self.fnn_distance * self.nn_high
				self.rCut_resolution = self.nn_resolution
		elif 'single_nn_measurement' in Predation_Information:
			if (('nn_low' in Predation_Information) or ('nn_high' in Predation_Information) or ('nn_resolution' in Predation_Information)):
				print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
				print('In Predation_Information, you have included single_nn_measurement as well as nn_low, nn_high, and/or nn_resolution.')
				print('You can only enter into Predation_Information either single_nn_measurement, or all of nn_low, nn_high, and nn_resolution')
				print('Only inlclude the following if you want to have a range of single_nn_measurement values:')
				print('\t* nn_low')
				print('\t* nn_high')
				print('\t* nn_resolution')
				print('Only include single_nn_measurement if you want to include only want to sample the similarity at one value of single_nn_measurement.')
				print('Predation_Information = '+str(Predation_Information))
				import pdb; pdb.set_trace()
				exit('This program will exit without completing')
			else:
				multiple_rCut = False; simple_rCut = True
				self.single_nn_measurement = float(Predation_Information['single_nn_measurement'])
				if not (1.0 <= single_nn_measurement <= 2.0):
					print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
					print('single_nn_measurement must be between 1.0 and 2.0')
					print('single_nn_measurement = '+str(single_nn_measurement))
					print('Check this')
					exit('This program will exit without completing')
				self.rCut = self.fnn_distance * self.single_nn_measurement
		else:
			print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
			print('In Predation_Information, you have or have not included single_nn_measurement as well as nn_low, nn_high, and/or nn_resolution.')
			print('You can only enter into Predation_Information either single_nn_measurement, or all of nn_low, nn_high, and nn_resolution')
			print('Only include the following if you want to have a range of nn values:')
			print('\t* nn_low')
			print('\t* nn_high')
			print('\t* nn_resolution')
			print('Only include single_nn_measurement if you want to include only want to sample the similarity at one value of rCut.')
			print('Predation_Information = '+str(Predation_Information))
			import pdb; pdb.set_trace()
			exit('This program will exit without completing')
	# lattice constant nearest neighbour distances have not been given. Using given rCut values
	else:
		if (('rCut_low' in Predation_Information) or ('rCut_high' in Predation_Information) or ('rCut_resolution' in Predation_Information)):
			if not (('rCut_low' in Predation_Information) and ('rCut_high' in Predation_Information) and ('rCut_resolution' in Predation_Information)):
				print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
				print('If you are wanting to have a range of rCut values, you need to include the following in Predation_Information:')
				print('\t* rCut_low')
				print('\t* rCut_high')
				print('\t* rCut_resolution')
				print('Make sure you have all of these in Predation_Information.')
				print('Predation_Information = '+str(Predation_Information))
				import pdb; pdb.set_trace()
				exit('This program will exit without completing')
			elif ('rCut' in Predation_Information):
				print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
				print('In Predation_Information, you have included rCut as well as rCut_low, rCut_high, and/or rCut_resolution.')
				print('You can only enter into Predation_Information either rCut, or all of Cut_low, rCut_high, and rCut_resolution')
				print('Only inlclude the following if you want to have a range of rCut values:')
				print('\t* rCut_low')
				print('\t* rCut_high')
				print('\t* rCut_resolution')
				print('Only include rCut if you want to include only want to sample the similarity at one value of rCut.')
				print('Predation_Information = '+str(Predation_Information))
				import pdb; pdb.set_trace()
				exit('This program will exit without completing')
			else:
				multiple_rCut = True; simple_rCut = False
				self.rCut_low = float(Predation_Information['rCut_low'])
				self.rCut_high = float(Predation_Information['rCut_high'])
				self.rCut_resolution = Predation_Information['rCut_resolution']
		elif ('rCut' in Predation_Information):	
			if (('rCut_low' in Predation_Information) or ('rCut_high' in Predation_Information) or ('rCut_resolution' in Predation_Information)):
				print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
				print('In Predation_Information, you have or or have not included rCut as well as rCut_low, rCut_high, and/or rCut_resolution.')
				print('You can only enter into Predation_Information either rCut, or all of Cut_low, rCut_high, and rCut_resolution')
				print('Only inlclude the following if you want to have a range of rCut values:')
				print('\t* rCut_low')
				print('\t* rCut_high')
				print('\t* rCut_resolution')
				print('Only include rCut if you want to include only want to sample the similarity at one value of rCut.')
				print('Predation_Information = '+str(Predation_Information))
				import pdb; pdb.set_trace()
				exit('This program will exit without completing')
			else:		
				multiple_rCut = False; simple_rCut = True
				self.rCut = float(Predation_Information['rCut'])
		else:
			print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
			print('In Predation_Information, you have included rCut as well as rCut_low, rCut_high, and/or rCut_resolution.')
			print('You can only enter into Predation_Information either rCut, or all of Cut_low, rCut_high, and rCut_resolution')
			print('Only inlclude the following if you want to have a range of rCut values:')
			print('\t* rCut_low')
			print('\t* rCut_high')
			print('\t* rCut_resolution')
			print('Only include rCut if you want to include only want to sample the similarity at one value of rCut.')
			print('Predation_Information = '+str(Predation_Information))
			import pdb; pdb.set_trace()
			exit('This program will exit without completing')
	# Make the set of rCuts
	if multiple_rCut:
		rCut_low = Predation_Information['rCut_low']
		rCut_high = Predation_Information['rCut_high']
		rCut_resolution = Predation_Information['rCut_resolution']
		rCuts = get_rCut_values(rCut_low,rCut_high,rCut_resolution)
	elif simple_rCut:
		rCut = Predation_Information['rCut']
		rCuts = [rCut]
	else:
		print('Error in creation of class SCM_Predation_Operator, in SCM_Predation_Operator.py')
		print('')
		import pdb; pdb.set_trace()
		exit()
	return rCuts # return rCuts to initialisation of SCM_Predation_Operator


	


from math import tanh, exp

class Fitness_Function:
	"""
	This class is designed to calculator the fitness value of a cluster, given a value of rho_i. 
	This class will include all the relavant information required, such as parameters, needed for the functions to convert rho_i into a fitness value.
	
	:param entries: This is a dictionary that contains all the input variables required for the fitness function. 
	:type  entries: dict.
	"""
	def __init__(self,**entries):
		self.__dict__.update(entries)
		if self.function == 'exponential':
			self.function = self.exponential_function
		elif self.function == 'tanh':
			self.function = self.tanh_function
		elif self.function == 'linear':
			self.linear_function = linear_function
		elif self.function == 'direct':
			self.linear_function = direct_function
		else:
			print('Error in initiating the Fitness_Function object.')
			print('The "function" input must be either (and including function variables):')
			print('    exponential (alpha)')
			print('    tanh')
			print('    linear (gradient,coefficent)')
			print('    direct_function')
			print('Correct for this and try running the genetic algorithm again.')
			exit ('This algorithm will exit without completing')
		self.check_parameters()

	def check_parameters(self):
		if self.function == 'exponential':
			self.function = self.exponential_function
			has_alpha = 'alpha' in self.__dict__.keys()
			if not (has_alpha):
				print('Error in class Fitness_Function, in Fitness_Function.py')
				print('To use the exponential fitness function, you must include variables for "alpha".')
				print('What you currently include:')
				print('    alpha: '+str(has_alpha))
				print('')
				print('Information about the fitness function: '+str(self.__dict__))
				print('Check this out.')
				exit('The genetic algorithm will exit without completing.')
		elif self.function == 'tanh':
			pass
		elif self.function == 'linear':
			has_gradient = 'gradient' in self.__dict__.keys()
			has_coefficient = 'coefficient' in self.__dict__.keys()
			if not (has_gradient and has_coefficient):
				print('Error in class Fitness_Function, in Fitness_Function.py')
				print('To use the linear fitness function, you must include variables for "gradient" and "coefficient".')
				print('What you currently include:')
				print('    gradient: '+str(has_gradient))
				print('    coefficient: '+str(has_coefficient))
				print('')
				print('information about the fitness function: '+str(self.__dict__))
				print('Check this out.')
				exit('The genetic algorithm will exit without completing.')
		elif self.function == 'direct':
			pass

	def get_fitness(self,rho_i):
		"""
		This def will give the fitness value for a given value of rho_i

		:param rho_i: a value
		:type  rho_i: float

		:returns: value
		:rtypes: float

		"""
		try:
			value = self.function(rho_i)
		except OverflowError as err:
			value = float('inf') 
		return value

	# ----------------------------------------------------------------------------------------------------------

	def exponential_function(self,rho_i):
		"""
		This definition is designed to return the result of the exp function for this genetic algorithm.

		Equation written as required in the genetic algorithm.

		:param rho_i: a value
		:type  rho_i: float
		:param alpha: a value
		:type  alpha: float

		:returns: value
		:rtypes: float

		"""
		try:
			value = exp(-1.0*float(self.alpha)*float(rho_i))
		except OverflowError as err:
			value = float('inf') 
		return value

	def tanh_function(self,rho_i):
		"""
		This definition is designed to return the result of the tanh function for this genetic algorithm.

		Equation written as required in the genetic algorithm.

		:param rho_i: a value
		:type  rho_i: float

		:returns: value
		:rtypes: float

		"""
		return 0.5*(1.0 - tanh(2.0*float(rho_i) - 1.0))

	def linear_function(self,rho_i):
		"""
		This definition is designed to return the result of the linear function for this genetic algorithm.

		Equation written as required in the genetic algorithm.

		:param rho_i: a value
		:type  rho_i: float
		:param gradient: a value
		:type  gradient: float
		:param coefficent: a value
		:type  coefficent: float

		:returns: value
		:rtypes: float

		"""
		try:
			value = self.gradient*float(rho_i) + self.coefficent
		except OverflowError as err:
			value = float('inf') 
		return value

	def direct_function(self,rho_i):
		"""
		This definition is designed to return the value for rho_i without any mathematical changes to it.

		Equation written as required in the genetic algorithm.

		:param rho_i: a value
		:type  rho_i: float

		:returns: rho_i
		:rtypes: float

		"""
		return rho_i

	# ----------------------------------------------------------------------------------------------------------


#!/usr/bin/python
'''
Did_Find_LES.py, Geoffrey Weal, 08/03/2019

This program will determine which of your genetic algorithm trials have completed up to a certain generation. 
'''
import os
from Organisms.Postprocessing_Programs.Did_Complete_Main import has_all_trials_finished

path = os.getcwd()

def did_Trial_finish_successfully(filepath):
    with open(filepath,'r') as DETAILS_FILE:
        for line in DETAILS_FILE:
            if line.startswith("The Genetic Algorithm finished successfully."):
                return True
    return False

Overall_Trials_to_check = {}

for dirpath, dirnames, filenames in os.walk(path):
    dirnames.sort()
    if any([(dirname.startswith('Trial') and dirname.replace('Trial','').isdigit()) for dirname in dirnames]):
        completed_successfully, completed_Trials, incomplete_Trials, at_generation = has_all_trials_finished(dirpath, dirnames)
        toString = ''
        toString += '******************************************************************************\n'
        toString += '******************************************************************************\n'
        toString += 'This set of genetic algorithm trials finished '+('SUCCESSFULLY' if completed_successfully else 'UNSUCCESSFULLY')+'.\n'
        toString += 'The following Trials in '+str(dirpath)+' completed or did not complete.\n'
        toString += 'Completed Trials: '+str(completed_Trials)+'\n'
        toString += 'Incomplete Trials: '+str(incomplete_Trials)+'\n'
        toString += 'There were '+str(len(incomplete_Trials))+' incomplete trials.\n'
        if not at_generation == []:
            toString += 'The current generation of the incomplete trials:\n'
            for trial_gen in at_generation:
                toString += str(trial_gen[0])+': Gen '+str(trial_gen[1])+'\n'
        toString += '******************************************************************************\n'
        toString += '******************************************************************************\n'
        print(toString)
        with open(dirpath+'/TrialsCompletionDetails.txt','w') as TrialsCompletionDetailsTXT:
            TrialsCompletionDetailsTXT.write(toString)
        dirnames[:] = []
        filenames[:] = []
        if not completed_successfully:
            Overall_Trials_to_check[dirpath] = incomplete_Trials

print('########################################################################')
print('########################################################################')
print('Details of Trials that did not complete:')
if len(Overall_Trials_to_check) == 0:
    print('ALL Trials completed SUCCESSFULLY')
else:
    for dirpath, incomplete_Trials in Overall_Trials_to_check.iteritems():
        print('path: '+str(dirpath)+'; Trials to Repeat: '+str(incomplete_Trials))
print('########################################################################')
print('########################################################################')



import os

lock_name = 'ga_running.lock'

def Lock_Check():
	if os.path.exists(lock_name):
		print('------------------------------------------------------------------------')
		print('Issue with Running the Genetic Algorithm.')
		print('The genetic algorithm has found the file "'+lock_name+'" before running the genetic algorithm.')
		print('This means that the user has tried to run this program while the genetic algorithm is already running.')
		print('Check that you are not already running this program. If you are not currently running this program, remove the "'+lock_name+'" from teh directory and continue.')
		print('If you had to stop this program without safely closing it, this file will have remained. If that is the case, remove "'+lock_name+'" and play on.')
		print('')
		print('-> Note that you can get rid of all the run files in subfolders if you run "remove_lock_files.py" script in the terminal. See manual for more information about the "remove_lock_files.py" script.')
		print('')
		print('The genetic algorithm will exit without having begun.')
		exit('------------------------------------------------------------------------')

def Lock_Set():
	if os.path.exists(lock_name):
		print('Issue with '+lock_name+'. Check programming of Lock.py')
		exit('The genetic algorithm will exit without having begun.')
	with open(lock_name,'w') as lock_file:
		lock_file.write('\n')

def Lock_Check_and_Set():
	Lock_Check()
	Lock_Set()

def Lock_Remove():
	if not os.path.exists(lock_name): 
		print('------------------------------------------------------------------------')
		print('Something weird had happened. The genetic algorithm is wanting to remove the lock file, but '+lock_name+' does not exist in the directory.')
		print('This is fine, but just note something weird has happened and may pay to figure out what happened.')
		print('Continuing the genetic algorithm, as the genetic algorithm will be finishing up')
		print('------------------------------------------------------------------------')
	else:
		os.remove(lock_name)


"""
ExternalDefinitions.py, 12/04/2017, Geoffrey R Weal

This python script is designed to hold subsidery definitions used by this Genetic Algorithm.

"""

from ase import Atom, Atoms
import copy, os
import collections

class atom_single_connection:
	"""
	This class is designed to record all the atoms an atom in a cluster is bonded/neighboured to, where a bond is
	defined as a atom-atom distance less than max_distance_between_atoms.
	:param atom: The dir of the atom in the cluster
	:type  atom: int
	"""
	def __init__(self,atom):
		self.atom = atom
		self.bonded_to = [] # The dir of the atoms that this self.atom is bonded to

def get_distance(atom1,atom2):
	xx = atom1[0] - atom2[0]
	yy = atom1[1] - atom2[1]
	zz = atom1[2] - atom2[2]
	return (xx*xx + yy*yy + zz*zz) ** 0.5

def Exploded(cluster, max_distance_between_atoms):
	"""
	This definition is designed to check to make sure a cluster has not exploded. This means that all the
	atoms in space are closely bound together as a nanoparticle rather than some atoms being
	disconnected from the majority of atoms in a cluster. This method works as follows:

	* Every atom is checked for neighbours/ other atoms it is bonded to in the cluster. The information of the neighbours of each atom are placed in a neighbour list
	* Write this later

	:param cluster: the cluster to check if all clusters are attached together.
	:type  cluster: ASE.Atoms
	:param max_distance_between_atoms: defines what the maximum length of a bond is in your cluster.
	:type  max_distance_between_atoms: float
	"""
	'''
	First the neighbours of all atoms in the cluster are recorded. These are all the atoms that have a bond length less than max_distance_between_atoms.
	'''
	# Initialise the list neighbours_list
	neighbours_list = []
	for ii in range(len(cluster)):
		neighbours_list.append(atom_single_connection(ii))
	# find and record the bonds for each atom in the cluster.
	cluster_positions = cluster.arrays['positions']
	for ii in range(len(neighbours_list)):
		for jj in range(ii+1,len(neighbours_list)): # from ii+1 to len(neighbours_list) as to prevent double recording of bonds.
			distance = get_distance(cluster_positions[neighbours_list[ii].atom],cluster_positions[neighbours_list[jj].atom]) # obtain the distance between two atoms
			# if distance < max_distance_between_atoms, the neighbours_list ii and jj are bonded together,
			# so record this in the instance of the class atom_single_connection for this atom
			if distance < max_distance_between_atoms:
				neighbours_list[ii].bonded_to.append(jj)
				neighbours_list[jj].bonded_to.append(ii)
	'''
	We now check to see that every atom is connect to a single entity (i.e. a cluster). This proceedure will
	make sure that the cluster is not in pieces.
	
	The algorithm works as follows:
	* Starting at atom 0, take a tree path through each bonding system to see if you can walk though to every atom in the cluster.
	* Get the neighbours of atom 0. these neighbours 
	* If this is not true, then the atom is in some way disconnected and so is not a true cluster.
	'''
	Cluster_paths = [0]
	atoms_not_reached = list(range(len(neighbours_list)))
	while not Cluster_paths == []:
		atom_to_explore = Cluster_paths.pop(0)
		if atom_to_explore in atoms_not_reached:
			atoms_not_reached.remove(atom_to_explore)
		Cluster_paths += neighbours_list[atom_to_explore].bonded_to
		while not neighbours_list[atom_to_explore].bonded_to == []:
			atoms_bonded_to = neighbours_list[atom_to_explore].bonded_to[0]
			neighbours_list[atoms_bonded_to].bonded_to.remove(atom_to_explore)
			neighbours_list[atom_to_explore].bonded_to.remove(atoms_bonded_to)

	if atoms_not_reached == []:
		return False
	else:
		return True

def InclusionRadiusOfCluster(cluster):
	"""
	Find the radius of a sphere that could completely enclose the cluster, with radius from the centre of mass
	to the most outer atom from the centre of mass.
	
	:param cluster: The cluster that the user would like to find the maximum radius from the centre of mass.
	:type  cluster: ASE.Atoms

	:returns: max(size): The radius of the cluster from the from the centre of mass to the most outer atom from the centre of mass (in Angstroms). This radius is for a sphere that will definitely enclose the cluster within.
	:rtype: float
	"""
	'''
	clus = copy.deepcopy(cluster)
	clus.center(vacuum=None, axis=(0, 1, 2), about=0.)
	size = []
	for atom in clus:
		dist = (atom.x*atom.x + atom.y*atom.y + atom.z*atom.z)**(1./2.)
		size.append(dist)
	'''
	max_size = -float('inf')
	for index1 in range(len(cluster)):
		for index2 in range(index1+1,len(cluster)):
			x_dist = cluster[index1].x - cluster[index2].x
			y_dist = cluster[index1].y - cluster[index2].y
			z_dist = cluster[index1].z - cluster[index2].z
			distance_between_atoms = (x_dist**2.0 + y_dist**2.0 + z_dist**2.0)**(0.5)
			if distance_between_atoms > max_size:
				max_size = distance_between_atoms
	return max_size/2.0

'''
def Rounding_Method(number,rounding_criteria):
	"""
	This definition will round the input number to the decimal place specified by rounding_criteria.
	The input number in the genetic algorithm is usually the energy of the cluster.

	:param number: The value that you would like to round (usually energy)
	:type  number: float
	:param rounding_criteria: The number of decimal places the value should be rounded to.
	:type  rounding_criteria: int
	
	"""
	round_by = (10**float(rounding_criteria))
	if rounding_criteria > 12:
		print('Error, the maximum decimal place rounding, to avoid numerical errors, is 12.')
		print('rounding_criteria: '+str(rounding_criteria))
		exit('Check this out. THis progra will finish without completing.')
	number = int(float(number)*round_by)/round_by
	return number
'''

def get_elemental_makeup(cluster):
	"""
	This gives a list which indicates the types of elements, and the number of those elements, in the cluster

	This method has been coped from the def get_elemental_makeup from class Cluster, in Cluster.py

	:returns: A list which indicates the types of elements, and the number of those elements, in the cluster
	:rtype: {str: int, ...} (old output was [[str.,int],...])
	"""
	elements = Counter(cluster.get_chemical_symbols())
	return elements
	'''
	if len(cluster) == 0:
		return [['',0]]
	elements = {}
	for atom in cluster:
		if atom.symbol not in elements:
			elements[atom.symbol] = 1
		else:
			elements[atom.symbol] += 1
	#elements = sorted([[element,number] for element,number in elements.iteritems()], key=lambda x:x[0], reverse=False)
	return elements
	'''

def isinstance_of_dict(dictionary,name_of_dictionary):
	if not isinstance(dictionary,dict):
		print('Error in def isinstance_of_dict/Difference_in_elements, in ExternalDefinitions.py')
		print('The input '+str(name_of_dictionary)+' is not a dictionary.')
		print(str(name_of_dictionary)+' = '+str(dictionary))
		print('Check this out.')
		import pdb; pdb.set_trace()
		exit('This program will finish without completing.')

def AtomInClusterPosition(atom,cluster):
	"""
	This definition will return true if their is an atom already at a particular position in the cluster.
	Prevent two atoms being in the exact same position which local optimisation techniques can not handle.

	Inputs:
		atom (ase.atom): The atom you want to add to the cluster, to check that atom will not be in the same position as any atom in the cluster.
		cluster (ase.atoms or GA.Cluster): The cluster that the atom is to be added to. 
	"""
	for atom_cluster in cluster:
		if atom.x == atom_cluster.x and atom.y == atom_cluster.y and atom.z == atom_cluster.z:
			return True
	return False

def is_position_already_occupied_by_an_atom_in_Cluster(atom_position,cluster,atom_indices_to_exclude_from_comparison=[]):
	"""
	This method determines if two atoms occupy the same position. 

	:param atom_position: The position of the atom, given as (x position, y position, z position). 
	:type  atom_position: (int, int, int)
	:param cluster: The cluster to investigate.
	:type  cluster: GA.Cluster
	:param atom_indices_to_exclude_from_comparison: list of the atom indices not to include in this analysis, THis should include the atom associated with the position list atom_position
	:type  atom_indices_to_exclude_from_comparison: list of int

	"""
	x_pos, y_pos, z_pos = atom_position
	atom_indices_to_check = [index for index in range(len(cluster)) if index not in atom_indices_to_exclude_from_comparison]
	for index in atom_indices_to_check:
		if x_pos == cluster[index].x and y_pos == cluster[index].y and z_pos == cluster[index].z:
			return True
	return False



#!/usr/bin/python
'''
Geoffrey Weal, remove_finish_files.py, 10/02/2021

This program is designed to remove all the finish files from subdirectories that this program is run from.
'''
import os, sys

path = os.getcwd()
for dirpath, dirs, files in os.walk(path):
	dirs.sort()
	if ('MDRun.py' in files and dirpath.split('/')[-1].startswith('Trial')) or ('run_eon.py' in files):
		if 'finish' in files:
			print('Removing finish file in '+str(dirpath))
			os.remove(dirpath+'/finish')
		dirs[:] = []
		files[:] = []
	if ('Run.py' in files and dirpath.split('/')[-1].startswith('Trial')):
		if 'finish' in files:
			print('Removing finish file in '+str(dirpath))
			os.remove(dirpath+'/finish')
		dirs[:] = []
		files[:] = []



def dot(v1, v2):
	return sum(x*y for x,y in zip(v1,v2))

from abc import ABC, abstractmethod
class Fitness_Operator(ABC):
	"""
	The Fitness class is designed to determine and assign the appropriate fitness value to clusters created during the genetic algorithm. 
	This class can be thought as an extension of the predation Operator. This class contains all the appropriate steps that are needed for 
	the fitnesses to be efficiency assigned to created clusters. 

	There are only two methods that need to be written for the predation Operator that is to be used. These are:
		* assign_population_fitnesses
		* assign_all_fitnesses

	Note: This class is intended to work as an interface only. Refer to the instruction manual as to how to use this interface to build your own predation Operator.

	:param fitness_information: The informatino needed by the Fitness_Operator
	:type  fitness_information: dict.
	:param population: The population to assign fitnesses to
	:type population: Organisms.GA.Population
	:param print_details: Print the details of the energy fitness operator. True if yes, False if no
	:type  print_details: bool
	"""
	def __init__(self, fitness_information, population, print_details):
		self.fitness_switch = fitness_information['Fitness Operator']
		self.fitness_information = fitness_information
		self.population = population
		self.print_details = print_details
		self.check()

	def check(self):
		"""
		This method will check that the self.fitness_switch is an available fitness operator. 
		"""
		if not self.fitness_switch in ['Energy','SCM + Energy']:
			error_string =  'Error in choosing Fitness Operators.\n'
			error_string += 'Predation_Operator must be either:\n'
			error_string += '\t"Energy" - The fitness of a cluster is based on its energy only.\n'
			error_string += '\t"SCM + Energy" - The fitness of a cluster is based on its energy and structural diversity as determined by the SCM.\n'
			error_string += 'Check this.\n'
			error_string += 'Fitness Operator = ' + str(fitness_switch)+'\n'
			#import pdb, traceback, sys
			#extype, value, tb = sys.exc_info()
			#traceback.print_exc()
			#pdb.post_mortem(tb)
			print(error_string,file=sys.stderr)
			import pdb; pdb.set_trace()
			exit()

	def __repr__(self):
		return str(self.__dict__)

	@abstractmethod
	def assign_initial_population_fitnesses(self):
		"""
		This method is designed to assign fitness values to the clusters of the population only at the start of the GA, when the population has been initialised

		This is an abstract method. 
		If you write your own diversiy and fitness classes, you MUST write something for this method.
		
		"""
		pass

	@abstractmethod
	def assign_resumed_population_fitnesses(self,resume_from_generation): 
		"""
		This method is designed to assign fitness values to the clusters of the population only at the beginning of a resumed GA. 

		This is an abstract method. 
		If you write your own diversiy and fitness classes, you MUST write something for this method.

		:param resume_from_generation: The current generation to resume from.
		:type  resume_from_generation: int
		
		"""
		pass

	@abstractmethod
	def assign_all_fitnesses_before_assess_against_predation_operator(self,all_offspring_pools,current_generation_no):
		"""
		This method is to be used in the GA program. 
		This will assign all the fitnesses of all clusters in the current generation (population and offspring) before the offspring are assessed to understand if they violate the predation scheme (i.e. an offspring is Class 1 similar to a cluster in the population, or another offspring).

		See the description given for the "assign_all_fitnesses" def on what the crux of this method is.

		If you write your own diversiy and fitness classes, you do not need to implement this method in your fitness class.

		:param all_offspring_pools: All of the offspring_pools
		:type  all_offspring_pools: list of Organisms.GA.Offspring_Pool
		:param current_generation_no: The current generation
		:type  current_generation_no: int
		"""
		pass

	@abstractmethod
	def assign_all_fitnesses_after_assess_against_predation_operator(self,all_offspring_pools,current_generation_no, offspring_to_remove):
		"""
		This method is to be used in the GA program. 
		This will assign all the fitnesses of all clusters in the current generation (population and offspring) after the offspring are assessed to understand if they violate the predation scheme (i.e. an offspring is Class 1 similar to a cluster in the population, or another offspring).

		See the description given for the "assign_all_fitnesses" def on what the crux of this method is.

		If you write your own diversiy and fitness classes, you do not need to implement this method in your fitness class.

		:param all_offspring_pools: All of the offspring_pools
		:type  all_offspring_pools: list of Organisms.GA.Offspring_Pool
		:param current_generation_no: The current generation
		:type  current_generation_no: int
		:param offspring_to_remove: a list of all the names of clusters in the offspring that will be removed.
		:type  offspring_to_remove: list of int
		"""
		pass

	@abstractmethod
	def assign_all_fitnesses_after_natural_selection(self,current_generation_no):
		"""
		This method is to be used in the GA program. 
		This will assign all the fitnesses to all the clusters in the population before the offspring are assessed to understand if they violate the predation operator (i.e. an offspring is Class 1 similar to a cluster in the population, or another offspring).

		See the description given for the "assign_population_fitnesses" def on what the crux of this method is.

		If you write your own diversiy and fitness classes, you do not need to implement this method in your fitness class.

		:param current_generation_no: The current generation
		:type  current_generation_no: int
		"""
		pass



metadata = {}
metadata['title'] = 'Genetic Algorithm Database'
metadata['default_columns'] = ['name', 'cluster_energy', 'id']
# ----------------------------------------------------
metadata['key_descriptions'] = {}
metadata['key_descriptions']['name'] = ('Name', 'Name of the cluster.', '')
metadata['key_descriptions']['gen_made'] = ('Generation Created', 'The generation the cluster was created.', '')
metadata['key_descriptions']['cluster_energy'] = ('Cluster Energy', 'Potential energy of the cluster.', 'energy_units')
metadata['key_descriptions']['ever_in_population'] = ('ever_in_population','This variable indicates if the cluster was ever in the population. If an offspring was made and was not ever accepted into the population, this variable will be False. If the cluster had been in the population for even one generation, this variable will be True.','bool')
metadata['key_descriptions']['excluded_because_violates_predation_operator'] = ('excluded_because_violates_predation_operator','This variable will determine if a cluster was excluded from the population because it violated the predation operator.','bool')
metadata['key_descriptions']['initial_population'] = ('initial_population','This variable will indicate if the cluster was apart of a newly created population.','bool')
metadata['key_descriptions']['removed_by_memory_operator'] = ('removed_by_memory_operator','This variable will indicate if a cluster is removed because it resembles a cluster in the memory operator.','bool')
# ----------------------------------------------------

title = metadata['title']
default_columns = metadata['default_columns']
key_descriptions = metadata['key_descriptions']



'''
Geoffrey Weal, MakeTrialsProgram.py, 10/02/2021

This program is designed to create all the Run.py files and submit.sl/mass_submit.sl that you desire.

This files are ordered into not subdirectories.
'''

import os, sys, inspect
from math import ceil, sqrt
from shutil import copyfile
from Organisms.Subsidiary_Programs.Create_submitSL_slurm_Main import make_submitSL, make_mass_submitSL_full, make_mass_submitSL_packets
 
def makeDirAndMove(folder):
    if not os.path.exists(str(folder)):
        os.makedirs(str(folder))
    os.chdir(str(folder))

def makeDir(folder):
    if not os.path.exists(str(folder)):
        os.makedirs(str(folder))

class MakeTrialsProgram:
    def __init__(self,cluster_makeup,pop_size,generations,no_offspring_per_generation,creating_offspring_mode,crossover_type,mutation_types,chance_of_mutation,r_ij,vacuum_to_add_length,Minimisation_Function,surface_details,epoch_settings,cell_length,memory_operator_information,predation_information,fitness_information,ga_recording_information,force_replace_pop_clusters_with_offspring,user_initialised_population_folder,rounding_criteria,print_details,no_of_cpus,dir_name,NoOfTrials,Condense_Single_Mention_Experiments,JobArraysDetails,making_files_for,finish_algorithm_if_found_cluster_energy=None,total_length_of_running_time=None,no_of_packets_to_make=None): 
        self.cluster_makeup=cluster_makeup
        self.pop_size=pop_size
        self.generations=generations
        self.no_offspring_per_generation=no_offspring_per_generation
        self.creating_offspring_mode=creating_offspring_mode
        self.crossover_type=crossover_type
        self.mutation_types=mutation_types
        self.chance_of_mutation=chance_of_mutation
        self.r_ij=r_ij
        self.vacuum_to_add_length=vacuum_to_add_length

        self.RunMinimisation_filename=os.path.relpath(inspect.getfile(Minimisation_Function)).replace('.py','').replace('/','.')
        self.Minimisation_Function_name=Minimisation_Function.__name__
        
        self.surface_details=surface_details
        self.epoch_settings=epoch_settings
        self.cell_length=cell_length
        self.memory_operator_information=memory_operator_information
        self.predation_information=predation_information
        self.fitness_information=fitness_information
        self.make_ga_recording_information_dict(ga_recording_information)
        self.force_replace_pop_clusters_with_offspring = force_replace_pop_clusters_with_offspring
        if user_initialised_population_folder in [None, 'None', 'none']:
            self.user_initialised_population_folder = None
        else:
            self.user_initialised_population_folder=os.path.abspath(user_initialised_population_folder)
        self.rounding_criteria=rounding_criteria
        self.print_details=print_details
        self.no_of_cpus=no_of_cpus
        self.finish_algorithm_if_found_cluster_energy=finish_algorithm_if_found_cluster_energy
        self.total_length_of_running_time=total_length_of_running_time

        self.dir_name=dir_name
        self.NoOfTrials=NoOfTrials
        self.making_files_for=making_files_for
        if self.making_files_for == 'slurm_JobArrays_packets':
            self.no_of_packets_to_make = no_of_packets_to_make
        self.Condense_Single_Mention_Experiments=Condense_Single_Mention_Experiments
        self.JobArraysDetails=JobArraysDetails
        if 'partition' not in self.JobArraysDetails:
            self.JobArraysDetails['partition'] = 'large'
        if 'python version' not in self.JobArraysDetails:
            self.JobArraysDetails['python version'] = 'Python/3.6.3-gimkl-2017a'
        self.check_inputs()
        self.MakeTrials()
        print('Created Multiple Trials files for Slurm')

    def make_ga_recording_information_dict(self,ga_recording_information):
        self.ga_recording_information = ga_recording_information
        if 'limit_number_of_clusters_recorded' in self.ga_recording_information and self.ga_recording_information['limit_number_of_clusters_recorded'] in ['inf', float('inf')]:
            self.ga_recording_information['limit_number_of_clusters_recorded'] = "inf"
        if 'limit_energy_height_of_clusters_recorded' in self.ga_recording_information and self.ga_recording_information['limit_number_of_clusters_recorded'] in ['inf', float('inf')]:
            self.ga_recording_information['limit_energy_height_of_clusters_recorded'] = "inf" #eV
        #import pdb; pdb.set_trace()

    def check_inputs(self):
        # Double check to make sure all things are in order.
        pass

    def MakeTrials(self):
        if self.making_files_for == 'individual':
            self.make_individual_trials()
        if self.making_files_for.startswith('slurm_JobArrays'):
            self.make_slurm_jobarray_files()

    def make_individual_trials(self):
        subdirname = self.get_folder_details_name()
        for trial_no in range(1,self.NoOfTrials+1):
            path_to = self.dir_name+'/'+subdirname+'/Trial'+str(trial_no)
            makeDir(path_to)
            #Write the Run.py file
            self.makeRunPY(path_to,self.cluster_makeup,self.pop_size,self.generations,self.no_offspring_per_generation,
                self.creating_offspring_mode,self.crossover_type,self.mutation_types,self.chance_of_mutation,self.r_ij,self.vacuum_to_add_length,
                self.RunMinimisation_filename,self.Minimisation_Function_name,self.surface_details,self.epoch_settings,self.cell_length,
                self.memory_operator_information,self.predation_information,self.fitness_information,self.ga_recording_information,
                self.force_replace_pop_clusters_with_offspring,self.user_initialised_population_folder,self.rounding_criteria,self.print_details,
                self.no_of_cpus,self.finish_algorithm_if_found_cluster_energy,self.total_length_of_running_time)
            # Copy the RunMinimisation.py file
            RunMinimisationPY_path = self.copyRunMinimisationPY(os.getcwd(), self.RunMinimisation_filename) # self.Minimisation_Function.__name__)
            self.PasteRunMinimisationPY(RunMinimisationPY_path, path_to, self.RunMinimisation_filename)
            # Make the submit.sl file for slurm
            make_submitSL(path_to,self.JobArraysDetails['project'],self.JobArraysDetails['time'],self.JobArraysDetails['nodes'],self.JobArraysDetails['ntasks_per_node'],self.JobArraysDetails['mem'],self.JobArraysDetails['partition'],self.JobArraysDetails['email'],self.JobArraysDetails['python version'])

    def make_slurm_jobarray_files(self):
        subdirname = self.get_folder_details_name()
        path_to = self.dir_name+'/'+subdirname
        makeDir(path_to)
        print('Making Run.py file in '+str(path_to))
        self.makeRunPY(path_to,self.cluster_makeup,self.pop_size,self.generations,self.no_offspring_per_generation,
            self.creating_offspring_mode,self.crossover_type,self.mutation_types,self.chance_of_mutation,self.r_ij,self.vacuum_to_add_length,
            self.RunMinimisation_filename,self.Minimisation_Function_name,self.surface_details,self.epoch_settings,self.cell_length,
            self.memory_operator_information,self.predation_information,self.fitness_information,self.ga_recording_information,
            self.force_replace_pop_clusters_with_offspring,self.user_initialised_population_folder,self.rounding_criteria,self.print_details,
            self.no_of_cpus,self.finish_algorithm_if_found_cluster_energy,self.total_length_of_running_time)
        # Copy the RunMinimisation.py file
        RunMinimisationPY_path = self.copyRunMinimisationPY(os.getcwd(), self.RunMinimisation_filename) # self.Minimisation_Function.__name__)
        self.PasteRunMinimisationPY(RunMinimisationPY_path, path_to, self.RunMinimisation_filename)
        # Make the submit.sl file for slurm
        if self.making_files_for == 'slurm_JobArrays_full':
            make_mass_submitSL_full(path_to,self.JobArraysDetails['project'],self.NoOfTrials,self.JobArraysDetails['time'],self.JobArraysDetails['nodes'],self.JobArraysDetails['ntasks_per_node'],self.JobArraysDetails['mem'],self.JobArraysDetails['partition'],self.JobArraysDetails['email'],self.RunMinimisation_filename,self.JobArraysDetails['python version'])
        elif self.making_files_for == 'slurm_JobArrays_packets':
            print('======================================================================')
            print('======================================================================')
            print()
            print('NOTE: YOU ARE USING slurm_JobArrays_packets')
            print()
            print('This will make slurm run a set of genetic algorithm trials within a single array')
            print('Only use this if your genetic algorithm will take less than 10 minutes to run')
            print('We have had issues if you run genetic algorithm trials where each GA takes over an hour to run with slurm_JobArrays_packets')
            print('Avoid this')
            print()
            print("You should only use this if you know your packet of GA trials will finish within the time you have given for JobArraysDetails['time'].")
            print()
            print('If each of your GA trials will take over an hour to complete, rerun this program but set making_files_for = "slurm_JobArrays_full"')
            print()
            print("Also Note, make sure you have set JobArraysDetails['time'] is set to as long as possible to avoid further issues.")
            print('======================================================================')
            print('======================================================================')
            make_mass_submitSL_packets(path_to,self.JobArraysDetails['project'],self.NoOfTrials,self.no_of_packets_to_make,self.JobArraysDetails['time'],self.JobArraysDetails['nodes'],self.JobArraysDetails['ntasks_per_node'],self.JobArraysDetails['mem'],self.JobArraysDetails['partition'],self.JobArraysDetails['email'],self.RunMinimisation_filename,self.JobArraysDetails['python version'])
        else:
            exit('Error, "making_files_for" must be either "slurm_JobArrays_full" or "slurm_JobArrays_packets". making_files_for = '+str(self.making_files_for))

    def get_folder_details_name(self):
        cluster_symbol_name = '_'.join(str(symbol) for symbol, number in self.cluster_makeup.items())
        cluster_makeup_name = ''.join((str(symbol)+str(number)) for symbol, number in self.cluster_makeup.items())
        if self.Condense_Single_Mention_Experiments:
            dirname = str(cluster_makeup_name)+'_P'+str(self.pop_size)+'_O'+str(self.no_offspring_per_generation)
        else:
            dirname = str(cluster_symbol_name)+'/'+str(cluster_makeup_name)+'/PopSize'+str(self.pop_size)+'/OffPerGenEquals'+str(self.no_offspring_per_generation)
        return dirname

    def makeRunPY(self,path_to,cluster_makeup,pop_size,generations,no_offspring_per_generation,creating_offspring_mode,crossover_type,mutation_types,chance_of_mutation,r_ij,vacuum_to_add_length,RunMinimisation_filename,Minimisation_Function_name,surface_details,epoch_settings,cell_length,memory_operator_information,predation_information,fitness_information,ga_recording_information,force_replace_pop_clusters_with_offspring,user_initialised_population_folder,rounding_criteria,print_details,no_of_cpus,finish_algorithm_if_found_cluster_energy,total_length_of_running_time):
        with open(path_to+'/Run.py','w+') as RunPY:
            print('from Organisms import GA_Program',file=RunPY)
            print('',file=RunPY)
            print('# This details the elemental and number of atom composition of cluster that the user would like to investigate',file=RunPY)
            print('',file=RunPY)
            print('cluster_makeup = '+str(cluster_makeup),file=RunPY)
            print('',file=RunPY)
            print('# Surface details',file=RunPY)
            print('surface_details = '+str(surface_details),file=RunPY)
            print('',file=RunPY)
            print('# These are the main variables of the genetic algorithm that with changes could affect the results of the Genetic Algorithm.',file=RunPY)
            print('pop_size = '+str(pop_size),file=RunPY)
            print('generations = '+str(generations),file=RunPY)
            print('no_offspring_per_generation = '+str(no_offspring_per_generation),file=RunPY)
            print('',file=RunPY)
            print('# These setting indicate how offspring should be made using the Mating and Mutation Proceedures',file=RunPY)
            print('creating_offspring_mode = "'+str(creating_offspring_mode)+'"',file=RunPY)
            print('crossover_type = "'+str(crossover_type)+'"',file=RunPY)
            print('mutation_types = '+str(mutation_types),file=RunPY)
            print('chance_of_mutation = '+str(chance_of_mutation),file=RunPY)
            print('',file=RunPY)
            print('# This parameter will tell the Organisms program if an epoch is desired, and how the user would like to proceed.',file=RunPY)
            print("epoch_settings = "+str(epoch_settings),file=RunPY)
            print('',file=RunPY)
            print('# These are variables used by the algorithm to make and place clusters in.',file=RunPY)
            print('r_ij = '+str(r_ij),file=RunPY)
            print('cell_length = '+str(cell_length),file=RunPY)
            print('vacuum_to_add_length = '+str(vacuum_to_add_length),file=RunPY)
            print('',file=RunPY)
            print('# The RunMinimisation.py algorithm is one set by the user. It contain the def Minimisation_Function',file=RunPY)
            print('# That is used for local optimisations. This can be written in whatever way the user wants to perform',file=RunPY)
            print('# the local optimisations. This is meant to be as free as possible.',file=RunPY)
            print('from '+str(RunMinimisation_filename)+' import '+str(Minimisation_Function_name),file=RunPY)
            print('',file=RunPY)
            print('# This dictionary includes the information required to prevent clusters being placed in the population if they are too similar to clusters in this memory_operator',file=RunPY)
            print('memory_operator_information = '+str(memory_operator_information),file=RunPY)
            print('',file=RunPY)
            print('# This dictionary includes the information required by the predation scheme.',file=RunPY)
            print('predation_information = '+str(predation_information),file=RunPY)
            print('',file=RunPY)
            print('# This dictionary includes the information required by the fitness scheme',file=RunPY)
            print('fitness_information = '+str(fitness_information),file=RunPY)
            print('',file=RunPY)
            print('# Variables required for the Recording_Cluster.py class/For recording the history as required of the genetic algorithm.',file=RunPY)
            print('ga_recording_information = '+str(ga_recording_information),file=RunPY)
            print('',file=RunPY)
            print('# These are last techinical points that the algorithm is designed in mind',file=RunPY)
            print('force_replace_pop_clusters_with_offspring = '+str(force_replace_pop_clusters_with_offspring),file=RunPY)
            if user_initialised_population_folder in [None, 'None', 'none']:
                print('user_initialised_population_folder = None',file=RunPY)
            else:
                print('user_initialised_population_folder = "'+str(user_initialised_population_folder)+'"',file=RunPY)
            print('rounding_criteria = '+str(rounding_criteria),file=RunPY)
            print('print_details = '+str(print_details),file=RunPY)
            print('no_of_cpus = '+str(no_of_cpus),file=RunPY)
            if finish_algorithm_if_found_cluster_energy:
                print('finish_algorithm_if_found_cluster_energy = '+str(finish_algorithm_if_found_cluster_energy),file=RunPY)
            else:
                print('finish_algorithm_if_found_cluster_energy = None',file=RunPY)
            if total_length_of_running_time:
                print('total_length_of_running_time = '+str(total_length_of_running_time),file=RunPY)
            else:
                print('total_length_of_running_time = None',file=RunPY)
            print('',file=RunPY)
            print("''' ---------------- '''",file=RunPY)
            print('GA_Program(cluster_makeup=cluster_makeup,',file=RunPY)
            print('    pop_size=pop_size,',file=RunPY)
            print('    generations=generations,',file=RunPY)
            print('    no_offspring_per_generation=no_offspring_per_generation,',file=RunPY)
            print('    creating_offspring_mode=creating_offspring_mode,',file=RunPY)
            print('    crossover_type=crossover_type,',file=RunPY)
            print('    mutation_types=mutation_types,',file=RunPY)
            print('    chance_of_mutation=chance_of_mutation,',file=RunPY)
            print('    r_ij=r_ij,',file=RunPY)
            print('    vacuum_to_add_length=vacuum_to_add_length,',file=RunPY)
            print('    Minimisation_Function=Minimisation_Function,',file=RunPY)
            print('    surface_details=surface_details,',file=RunPY)
            print('    epoch_settings=epoch_settings,',file=RunPY)
            print('    cell_length=cell_length,',file=RunPY)
            print('    memory_operator_information=memory_operator_information,',file=RunPY)
            print('    predation_information=predation_information,',file=RunPY)
            print('    fitness_information=fitness_information,',file=RunPY)
            print('    ga_recording_information=ga_recording_information,',file=RunPY)
            print('    force_replace_pop_clusters_with_offspring=force_replace_pop_clusters_with_offspring,',file=RunPY)
            print('    user_initialised_population_folder=user_initialised_population_folder,',file=RunPY)
            print('    rounding_criteria=rounding_criteria,',file=RunPY)
            print('    print_details=print_details,',file=RunPY)
            if not finish_algorithm_if_found_cluster_energy:
                print('    no_of_cpus=no_of_cpus,',file=RunPY)
                print('    total_length_of_running_time=total_length_of_running_time)',file=RunPY)
            else:
                print('    no_of_cpus=no_of_cpus,',file=RunPY)
                print('    finish_algorithm_if_found_cluster_energy=finish_algorithm_if_found_cluster_energy,',file=RunPY)
                print('    total_length_of_running_time=total_length_of_running_time)',file=RunPY)
            print("''' ---------------- '''",file=RunPY)

    ##########################################################################################################

    def copyRunMinimisationPY(self, path_to, RunMinimisation_Name):
        RunMinimisationPY_path = path_to+'/'+RunMinimisation_Name+'.py'
        return RunMinimisationPY_path

    def PasteRunMinimisationPY(self,RunMinimisationPY_path,path_to,RunMinimisation_Name=None):
        if RunMinimisation_Name == None:
            RunMinimisation_Path = os.path.basename(os.path.normpath(RunMinimisationPY_path))
        else:
            RunMinimisation_Path = path_to+'/'+RunMinimisation_Name+'.py'
        copyfile(RunMinimisationPY_path,os.getcwd()+'/'+RunMinimisation_Path)

    ##########################################################################################################

    def writeDetails(self,cluster_makeup,nPool,generations,noOffspringPerGeneration,Creating_Offspring_Mode,crossType,ChanceOfMutation,mutTypes,r_ij,boxtoplaceinlength,vacuumAdd,RunMinimisation_Name,Predation_Information,recording_cluster_scheme_information,write_folders,write_Initial_Population_folder,user_initialised_population_folder,rounding_criteria):
        with open('Genetic_Algorithm_Run_Information.txt','w') as InfoTXT:
            InfoTXT.write('===========================================================================================\n')
            InfoTXT.write('===========================================================================================\n')
            InfoTXT.write('Here provides all the information about this set of repeated Genetic Algorithm experiments.\n')
            InfoTXT.write('===========================================================================================\n')
            InfoTXT.write('===========================================================================================\n')
            InfoTXT.write('\n')
            InfoTXT.write('-----------------\n')
            InfoTXT.write('Makeup of Cluster\n')
            InfoTXT.write('-----------------\n')
            InfoTXT.write('This details the elemental and number of atom composition of cluster that the user would like to investigate\n')
            InfoTXT.write('\n')
            InfoTXT.write('Atomic Makeup of the Cluster: '+str(cluster_makeup)+'\n')
            InfoTXT.write('\n')
            InfoTXT.write('-------------------------------------------\n')
            InfoTXT.write('General Details about the Genetic Algorithm\n')
            InfoTXT.write('-------------------------------------------\n')
            InfoTXT.write('These are the main variables of the genetic algorithm that with changes could affect the results of the Genetic Algorithm.\n')
            InfoTXT.write('\n')
            InfoTXT.write('Population Size: '+str(nPool)+'\n')
            InfoTXT.write('Number of Generations: '+str(generations)+'\n')
            InfoTXT.write('Number of Offspring Generated per Generation: '+str(noOffspringPerGeneration)+'\n')
            InfoTXT.write('\n')
            InfoTXT.write('-------------------------------------------------\n')
            InfoTXT.write('Details about the Mating and Mutation Procedures\n')
            InfoTXT.write('-------------------------------------------------\n')
            InfoTXT.write('These setting indicate how offspring should be made using the Mating and Mutation Procedures\n')
            InfoTXT.write('\n')
            InfoTXT.write('Mate and/or Mutate? (both/either): '+str(Creating_Offspring_Mode)+'\n')
            InfoTXT.write('Mating Procedure: '+str(crossType)+'\n')
            InfoTXT.write('Mutation Procedure (and chance they will occur out of 1): '+str(mutTypes)+'\n')
            InfoTXT.write('Chance of a Mutation Occurring: '+str(ChanceOfMutation)+' ('+str(float(ChanceOfMutation)*100.0)+' %)'+'\n')
            InfoTXT.write('\n')
            InfoTXT.write('-------------------------------------------------\n')
            InfoTXT.write('Other General Details about the Genetic Algorithm\n')
            InfoTXT.write('-------------------------------------------------\n')
            InfoTXT.write('These are variables used by the algorithm to make and place clusters in.\n')
            InfoTXT.write('\n')
            InfoTXT.write('r_ij: '+str(r_ij)+'\n')
            InfoTXT.write('boxtoplaceinlength: '+str(boxtoplaceinlength)+'\n')
            InfoTXT.write('vacuumAdd (vacuum to add around cluster): '+str(vacuumAdd)+'\n')
            InfoTXT.write('\n')
            InfoTXT.write('-------------------------\n')
            InfoTXT.write('Local Minimisation Script\n')
            InfoTXT.write('-------------------------\n')
            InfoTXT.write('The RunMinimisation.py algorithm is one set by the user. It contain the def Minimisation_Function\n')
            InfoTXT.write('That is used for local optimisations. This can be written in whatever way the user wants to perform\n')
            InfoTXT.write('the local optimisations. This is meant to be as free as possible.\n')
            InfoTXT.write('\n')
            InfoTXT.write('Local Minimisation Script Name: '+str(RunMinimisation_Name)+'\n')
            InfoTXT.write('\n')
            InfoTXT.write('----------------------------\n')
            InfoTXT.write('Predation Scheme Information\n')
            InfoTXT.write('----------------------------\n')
            InfoTXT.write('This switch tells the genetic algorithm the type of predation scheme they want to place on the genetic algoithm.\n')
            InfoTXT.write('\n')
            for Predation_entry in Predation_Information:
                InfoTXT.write(str(Predation_entry)+': '+str(Predation_Information[Predation_entry])+'\n')
            InfoTXT.write('\n')
            InfoTXT.write('------------------------------\n')
            InfoTXT.write('Recording Clusters Information\n')
            InfoTXT.write('------------------------------\n')
            InfoTXT.write('Place all the recording_cluster_scheme_information variables together into a list as below\n')
            InfoTXT.write('\n')
            for recording_cluster_entry in recording_cluster_scheme_information:
                InfoTXT.write(str(recording_cluster_entry)+': '+str(recording_cluster_scheme_information[recording_cluster_entry])+'\n')
            InfoTXT.write('\n')
            InfoTXT.write('--------------\n')
            InfoTXT.write('Other Settings\n')
            InfoTXT.write('--------------\n')
            InfoTXT.write('These are last techinical points that the algorithm is designed in mind\n')
            InfoTXT.write('\n')
            InfoTXT.write('write_folders: '+str(write_folders)+'\n')
            InfoTXT.write('write_Initial_Population_folder: '+str(write_Initial_Population_folder)+'\n')
            InfoTXT.write('user_initialised_population_folder: '+str(user_initialised_population_folder)+'\n')
            InfoTXT.write('rounding_criteria: '+str(rounding_criteria)+'\n')



import os
import math
from Organisms.GA.Lock import Lock_Remove

def if_to_finish_because_found_cluster_energy(self, all_energies_of_offpsring):
	if not self.finish_algorithm_if_found_cluster_energy == None:
		for cluster in self.population:
			if round(cluster.energy,self.finish_algorithm_if_found_cluster_rounding) == self.finish_algorithm_if_found_cluster_energy:
				self.energyprofile.add_found_LES_note()
				return True
		for offspring_energy in all_energies_of_offpsring:
			if round(offspring_energy,self.finish_algorithm_if_found_cluster_rounding) == self.finish_algorithm_if_found_cluster_energy:
				self.energyprofile.add_found_LES_note()
				return True
	return False

def check_for_duplicates(collection_cluster_names):
	if len(collection_cluster_names) == len(set(collection_cluster_names)):
		return False
	else:
		return True

def check_names_1(population):	
	pop_cluster_names = population.get_cluster_names()
	if check_for_duplicates(pop_cluster_names):
		print('Error in def check_names, in GA_Program.py')
		print('We have found duplicates in both the population')
		print('pop_cluster_names: '+str(pop_cluster_names.sort()))
		print('Check this out')
		Lock_Remove()
		raise Exception('Program finishing without completing')

def check_names_2(population,offspring_pool):
	pop_cluster_names = population.get_cluster_names()
	off_cluster_names = offspring_pool.get_cluster_names()
	if check_for_duplicates(off_cluster_names):
		print('Error in def check_names, in GA_Program.py')
		print('We have found duplicates in both the population')
		print('off_cluster_names: '+str(off_cluster_names.sort()))
		print('Check this out')
		Lock_Remove()
		raise Exception('Program finishing without completing')
	check = any(cluster_name in off_cluster_names for cluster_name in pop_cluster_names)
	if check:
		print('Error in def check_names, in GA_Program.py')
		print('We have found the cluster with the same name in both the offspring and population')
		print('pop_cluster_names: '+str(pop_cluster_names))
		print('off_cluster_names: '+str(off_cluster_names))
		duplicate_clusters = []
		for cluster_name in pop_cluster_names:
			if cluster_name in off_cluster_names:
				duplicate_clusters.append(cluster_name)
		print('The cluster names in the population and offspring: '+str(duplicate_clusters))
		print('Check this out')
		Lock_Remove()
		raise Exception('Program finishing without completing')

def remove_cluster_files(self):
	"""
	Will remove the xyz or database files when the Organisms program is finished. However, it is recommended that you keep the population database or xyz files when the Organisms program finishes, just in case you would like to extend the Organisms program to perform more generations.
	"""
	if self.remove_cluster_files_at_end:
		self.population.delete_collection()

def add_metadata(self):
	"""
	This is included as this seems to not be added to the collections database when it is first created. Fixing an issue with ASE.
	"""
	self.ga_recording_system.add_metadata()
	self.population.add_metadata()
	#self.offspring_pool.add_metadata()

def floor_float(number,decimal_place):
	number_up = 10.0 ** float(decimal_place)
	return math.floor(float(number) * number_up)/number_up

def reset_population(self,generation_number):
	population_reset_settings = (self.cluster_makeup, self.surface, self.Minimisation_Function, self.memory_operator, self.predation_operator, self.fitness_operator, self.epoch, self.cell_length, self.vacuum_to_add_length, self.r_ij, self.rounding_criteria, self.no_of_cpus, self.previous_cluster_name)
	epoch_due_to_population_energy_convergence = not self.population.is_there_an_energy_range(self.rounding_criteria)
	self.previous_cluster_name = self.epoch.perform_epoch(generation_number,self.population,self.energyprofile,population_reset_settings,epoch_due_to_population_energy_convergence=epoch_due_to_population_energy_convergence)
	self.ga_recording_system.add_collection(self.population, []) # save this new reset population

import time, sys
# update_progress() : Displays or updates a console progress bar
## Accepts a float between 0 and 1. Any int will be converted to a float.
## A value under 0 represents a 'halt'.
## A value at 1 or bigger represents 100%
def update_progress(progress):
    barLength = 50 # Modify this to change the length of the progress bar
    status = ""
    if isinstance(progress, int):
        progress = float(progress)
    if not isinstance(progress, float):
        progress = 0
        status = "error: progress var must be float\r\n"
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(barLength*progress))
    text = "\rPercent: [{0}] {1}% {2}".format( "#"*block + "-"*(barLength-block), round(progress*100.0,1), status)
    sys.stderr.write(text)
    sys.stderr.flush()

def convert_seconds_to_normal_time(seconds):
    mins, secs = divmod(seconds, 60)
    return str(mins)+' mins and '+str(secs)+' secs'

def check_files_for_readable_and_writable():
	not_readable_and_writable_files = []
	not_readable_and_writable_folders = []
	root_path = os.path.dirname(__file__)
	files_to_check = ['epoch_data','GA_Run_Details.txt']
	folders_to_check = ['Population','Recorded_Data']
	# check root GA files
	for name in files_to_check:
		path_to_file = name #os.path.relpath(os.path.join('.', name))
		if not os.access(path_to_file,0o777):
			not_readable_and_writable_files.append(path_to_file)
	# check root GA folders
	for folder_to_check in folders_to_check:
		for root, dirs, files in os.walk(folder_to_check, topdown=False):
			for name in files:
				path_to_file = os.path.relpath(os.path.join(root, name))
				if not os.access(path_to_file,0o777):
					not_readable_and_writable_files.append(path_to_file)
			for name in dirs:
				path_to_folder = os.path.relpath(os.path.join(root, name))
				if not os.access(path_to_folder,0o777):
					not_readable_and_writable_folders.append(path_to_folder)
	# turning folders to chmod 777
	if not len(not_readable_and_writable_files) == 0 or len(not_readable_and_writable_folders) == 0:
		try:
			for path_to_file in not_readable_and_writable_files:
				os.access(path_to_file,0o777)
			for path_to_folder in not_readable_and_writable_folders:
				os.access(path_to_folder,0o777)
		except:
			print('--------------------------------------------------------------------------------------')
			print('Error in your genetic algorithm program')
			print('Some of your files and folders are not both readable and writable.')
			print('All of your files in your genetic algorithm need to be readable and writable by anyone')
			print('files and folders that you need to check')
			print('-----------------')
			print('Files')
			for path_to_file in not_readable_and_writable_files:
				print('\t'+str(path_to_file))
			print('-----------------')
			print('Folders')
			for path_to_folder in not_readable_and_writable_folders:
				print('\t'+str(path_to_folder))
			print('-----------------')
			print("You can attempt to change the permissions of these files by performing the following in this genetic algorithm run's folder in the terminal:")
			print('\tchmod -R 777 *')
			print('If this does not work, this likely means that this genetic algorithm was run last by someone else.')
			print('Check this out and talk to the person who last ran this genetic algorithm run.')
			exit('This program will finish without completing.')





import os, sys, time

from Organisms import __version__

def get_version_number():
	version = __version__
	return version

def version_no():
	"""
	Will provide the version of the Organisms program
	"""
	path = __file__
	version = get_version_number() 
	return version

def GA_Program_Logo():
	"""
	Provides the opening logo for the Organisms program.
	"""
	print("")
	print(".--------------------------------------------------------------------------.")
	print("|                             ,                                            |")
	print("|              ,_     ,     .'<_                                           |")
	print("|             _> `'-,'(__.-' __<                                           |")
	print("|             >_.--(.. )  =;`                                   _          |")
	print("|                  `V-'`'\\/``                                  ('>         |")
	print("|                                                              /))@@@@@.   |")
	print('|         .----------------------------------.                /@"@@@@@()@  |')
	print("|         | Welcome to the Organisms program |               .@@()@@()@@@@ |")
	print("|         '----------------------------------'               @@@O@@@@()@@@ |")
	print("|                  ^                                         @()@@\\@@@()@@ |")
	print("|  _     (\\_/)     |                                          @()@||@@@@@' |")
	print("| ( \\    (0.o) ----'                 _                         '@@||@@@'   |")
	print("|  ) )   (> <)                     _(_)_            wWWWw   _     ||       |")
	print("| ( (  .-'''''-.  A.-.A      @@@@ (_)@(_)       _   (___) _(_)_   ||       |")
	print("|  \\ \\/         \\/ , , \\    @@()@@  (_)\\      _(_)_   Y  (_)@(_)  ||       |")
	print("|   \\   \\        ;= t =/     @@@@      `|/   (_)@(_) \\|/   (_)\\   ||       |")
	print("|    \\   |'''''-   ,--'       /        \\|     /(_)    |/      |   ||       |")
	print("|     / //     | ||        \\ |          | / \\|/      \\|      \\|/  ||       |")
	print("|    /_,))     |_'))       \\\\|//     \\\\\\|// \\|///   \\\\|//  \\\\\\|// ||       |")
	print("|^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^|")
	print("'--------------------------------------------------------------------------'")
	print("")

def Introducing_Remarks(self):
	"""
	Provides information about the settings for your Organisms run.
	"""
	line_length = 60
	print('#'*line_length)
	print('#'*line_length)
	print('#'*line_length)
	print('The Otago Research Genetic Algorithm for Nanoclusters, Including Structural Methods and Similarity (Organisms) Program')
	GA_Program_Logo()
	print('Version: '+str(version_no()))
	print('#'*line_length)
	print('#'*line_length)
	time.sleep(1)
	print('')
	print('This algorithm is designed to perform a genetic algorithm to obtain low energetic structures for a given cluster composition.')
	print('')
	print('This genetic algorithm will proceed with the following input parameters')
	print('')
	print('# This details the elemental and number of atom composition of cluster that the user would like to investigate')
	print('cluster_makeup = '+str(self.cluster_makeup))
	print('')
	print('# This details the information used to perform a global optimisation upon a chosen surface')
	print('surface_details = '+str(self.surface_details))
	print('')
	print('# These are the main variables of the genetic algorithm that with changes could affect the results of the Genetic Algorithm.')
	print('pop_size = '+str(self.pop_size))
	print('generations = '+str(self.generations))
	print('no_offspring_per_generation = '+str(self.no_offspring_per_generation))
	print('')
	print('# These setting indicate how offspring should be made using the Mating and Mutation Proceedures')
	print('creating_offspring_mode = '+str(self.creating_offspring_mode))
	print('crossover_type = '+str(self.crossover_type))
	print('mutation_types = '+str(self.mutation_types))
	print('mutation_types = '+str(self.chance_of_mutation))
	print('')
	print('# These are variables used by the algorithm to make and place clusters in.')
	print('r_ij = '+str(self.r_ij)+' Angstroms (This is the maximum possible bond length between two atoms in the cluster)')
	print('cell_length = '+str(self.cell_length)+' Angstroms')
	print('vacuum_to_add_length = '+str(self.vacuum_to_add_length)+' Angstroms')
	print('')
	print('# The RunMinimisation.py algorithm is one set by the user. It contain the def Minimisation_Function')
	print('# That is used for local optimisations. This can be written in whatever way the user wants to perform')
	print('# the local optimisations. This is meant to be as free as possible.')
	print('Minimisation_Function = '+str(self.Minimisation_Function))
	Minimisation_Function_script = self.Minimisation_Function.__module__
	Minimisation_Function_script_location = os.path.dirname(sys.modules[Minimisation_Function_script].__file__)
	print('Path of Minimisation_Function: '+str(Minimisation_Function_script_location+'/'+Minimisation_Function_script+'.py'))
	print('')
	print('# This details the information about the memory operator used in this genetic algorithm.')
	print('memory_operator = '+str(self.memory_operator))
	print('')
	print('# This details the information about the predation operator used in this genetic algorithm.')
	print('predation_operator = '+str(self.predation_operator))
	print('')
	print('# This details the information about the fitness operator used in this genetic algorithm.')
	print('fitness_operator = '+str(self.fitness_operator))
	print('')
	print('# This parameter will tell the Organisms program if an epoch is desired, and how the user would like to proceed.')
	print('epoch_settings: '+str(self.epoch))
	print('')
	print('# Variables required for the GA_Recording_System.py class/For recording the history as required of the genetic algorithm.')
	print('ga_recording_system = '+str(self.ga_recording_system))
	print('')
	print('# These are last techinical points that the algorithm is designed in mind')
	print('force_replace_pop_clusters_with_offspring = '+str(self.force_replace_pop_clusters_with_offspring))
	print('user_initialised_population_folder = '+str(self.user_initialised_population_folder))
	print('rounding_criteria = '+str(self.rounding_criteria))
	print('print_details = '+str(self.print_details))
	print('no_of_cpus = '+str(self.no_of_cpus))
	print('finish_algorithm_if_found_cluster_energy = '+str(self.finish_algorithm_if_found_cluster_energy))
	print('total_length_of_running_time: '+str(self.timer.get_total_length_of_running_time()))
	print('')
	print('#'*line_length)
	print('#'*line_length)
	print('#'*line_length)